from nbodykit.base.mesh import MeshSource
from nbodykit.base.catalog import CatalogSource, CatalogSourceBase
from nbodykit import _global_options
import numpy
import logging
import warnings

# for converting from particle to mesh
from pmesh import window
from pmesh.pm import RealField, ComplexField

class CatalogMesh(CatalogSource, MeshSource):
    """
    A view of a CatalogSource object which knows how to create a MeshSource
    object from itself.

    The original CatalogSource object is stored as the :attr:`base` attribute.

    Parameters
    ----------
    source : CatalogSource
        the input catalog that we are viewing as a mesh
    BoxSize :
        the size of the box
    Nmesh : int, 3-vector
        the number of cells per mesh side
    dtype : str
        the data type of the values stored on mesh
    weight : str
        column in ``source`` that specifies the weight value for each
        particle in the ``source`` to use when gridding
    value : str
        column in ``source`` that specifies the field value for each particle;
        the mesh stores a weighted average of this column
    selection : str
        column in ``source`` that selects the subset of particles to grid
        to the mesh
    position : str, optional
        column in ``source`` specifying the position coordinates; default
        is ``Position``
    interlaced : bool, optional
        use the interlacing technique of Sefusatti et al. 2015 to reduce
        the effects of aliasing on Fourier space quantities computed
        from the mesh
    compensated : bool, optional
        whether to correct for the window introduced by the grid
        interpolation scheme
    window : str, optional
        the string specifying which window interpolation scheme to use;
        see ``pmesh.window.methods``
    """
    logger = logging.getLogger('CatalogMesh')

    def __repr__(self):
        if isinstance(self.base, CatalogMesh):
            return repr(self.base)
        else:
            return "(%s as CatalogMesh)" % repr(self.base)

    def __new__(cls, source, BoxSize, Nmesh, dtype, weight,
                    value, selection, position='Position', interlaced=False,
                    compensated=False, window='cic', **kwargs):

        # source here must be a CatalogSource
        assert isinstance(source, CatalogSourceBase)

        # new, empty CatalogSource
        obj = CatalogSourceBase.__new__(cls, source.comm)

        # copy over size from the CatalogSource
        obj._size = source.size
        obj._csize = source.csize

        # copy over the necessary meta-data to attrs
        obj.attrs['BoxSize'] = BoxSize
        obj.attrs['Nmesh'] = Nmesh
        obj.attrs['interlaced'] = interlaced
        obj.attrs['compensated'] = compensated
        obj.attrs['window'] = window

        # copy meta-data from source too
        obj.attrs.update(source.attrs)

        # store others as straight attributes
        obj.dtype = dtype
        obj.weight = weight
        obj.value = value
        obj.selection = selection
        obj.position = position

        # add in the Mesh Source attributes
        MeshSource.__init__(obj, obj.comm, Nmesh, BoxSize, dtype)

        # finally set the base as the input CatalogSource
        # NOTE: set this AFTER MeshSource.__init__()
        obj.base = source

        return obj

    def gslice(self, start, stop, end=1, redistribute=True):
        """
        Execute a global slice of a CatalogMesh.

        .. note::
            After the global slice is performed, the data is scattered
            evenly across all ranks.

        As CatalogMesh objects are views of a CatalogSource, this simply
        globally slices the underlying CatalogSource.

        Parameters
        ----------
        start : int
            the start index of the global slice
        stop : int
            the stop index of the global slice
        step : int, optional
            the default step size of the global size
        redistribute : bool, optional
            if ``True``, evenly re-distribute the sliced data across all
            ranks, otherwise just return any local data part of the global
            slice
        """
        # sort the base object
        newbase = self.base.gslice(start, stop, end=end, redistribute=redistribute)

        # view this base class as a CatalogMesh (with default CatalogMesh parameters)
        toret = newbase.view(self.__class__)

        # attach the meta-data from self to returned sliced CatalogMesh
        return toret.__finalize__(self)

    def sort(self, keys, reverse=False, usecols=None):
        """
        Sort the CatalogMesh object globally across all MPI ranks
        in ascending order by the input keys.

        Sort columns must be floating or integer type.

        As CatalogMesh objects are views of a CatalogSource, this simply
        sorts the underlying CatalogSource.

        Parameters
        ----------
        *keys :
            the names of columns to sort by. If multiple columns are provided,
            the data is sorted consecutively in the order provided
        reverse : bool, optional
            if ``True``, perform descending sort operations
        usecols : list, optional
            the name of the columns to include in the returned CatalogSource
        """
        # sort the base object
        newbase = self.base.sort(keys, reverse=reverse, usecols=usecols)

        # view this base class as a CatalogMesh (with default CatalogMesh parameters)
        toret = newbase.view(self.__class__)

        # attach the meta-data from self to returned sliced CatalogMesh
        return toret.__finalize__(self)

    def __slice__(self, index):
        """
        Return a slice of a CatalogMesh object.

        This slices the CatalogSource object stored as the :attr:`base`
        attribute, and then views that sliced object as a CatalogMesh.

        Parameters
        ----------
        index : array_like
            either a dask or numpy boolean array; this determines which
            rows are included in the returned object

        Returns
        -------
        subset
            the particle source with the same meta-data as ``self``, and
            with the sliced data arrays
        """
        # this slice of the CatalogSource will be the base of the mesh
        base = super(CatalogMesh, self).__slice__(index)

        # view this base class as a CatalogMesh (with default CatalogMesh parameters)
        toret = base.view(self.__class__)

        # attach the meta-data from self to returned sliced CatalogMesh
        return toret.__finalize__(self)

    def copy(self):
        """
        Return a shallow copy of ``self``.

        .. note::
            No copy of data is made.

        Returns
        -------
        CatalogMesh :
            a new CatalogMesh that holds all of the data columns of ``self``
        """
        # copy the base and view it as a CatalogMesh
        toret = self.base.copy().view(self.__class__)

        # attach the meta-data from self to returned sliced CatalogMesh
        return toret.__finalize__(self)

    def __finalize__(self, other):
        """
        Finalize the creation of a CatalogMesh object by copying over
        attributes from a second CatalogMesh.

        This also copies over the relevant MeshSource attributes via a
        call to :func:`MeshSource.__finalize__`.

        Parameters
        ----------
        obj : CatalogMesh
            the second CatalogMesh to copy over attributes from
        """
        if isinstance(other, CatalogSourceBase):
            self = CatalogSourceBase.__finalize__(self, other)

        if isinstance(other, MeshSource):
            self = MeshSource.__finalize__(self, other)

        return self

    @property
    def interlaced(self):
        """
        Whether to use interlacing when interpolating the density field.
        See :ref:`the documentation <interlacing>` for further details.

        See also: Section 3.1 of
        `Sefusatti et al. 2015 <https://arxiv.org/abs/1512.07295>`_
        """
        return self.attrs['interlaced']

    @interlaced.setter
    def interlaced(self, interlaced):
        self.attrs['interlaced'] = interlaced

    @property
    def window(self):
        """
        String specifying the name of the interpolation kernel when
        gridding the density field.

        See :ref:`the documentation <window-kernel>` for further details.

        .. note::
            Valid values must be in :attr:`pmesh.window.methods`
        """
        return self.attrs['window']

    @window.setter
    def window(self, value):
        assert value in window.methods
        self.attrs['window'] = value.lower() # lower to compare with compensation

    @property
    def compensated(self):
        """
        Boolean flag to indicate whether to correct for the windowing
        kernel introduced when interpolating the discrete particles to
        a continuous field.

        See :ref:`the documentation <compensation>` for further details.
        """
        return self.attrs['compensated']

    @compensated.setter
    def compensated(self, value):
        self.attrs['compensated'] = value

    def to_real_field(self, out=None, normalize=True):
        """
        Paint the density field, by interpolating the position column
        on to the mesh.

        This computes the following meta-data attributes in the process of
        painting, returned in the :attr:`attrs` attributes of the returned
        RealField object:

        - N : int
            the (unweighted) total number of objects painted to the mesh
        - W : float
            the weighted number of total objects, equal to the collective
            sum of the 'weight' column
        - shotnoise : float
            the Poisson shot noise, equal to the volume divided by ``N``
        - num_per_cell : float
            the mean number of weighted objects per cell

        .. note::

            The density field on the mesh is normalized as :math:`1+\delta`,
            such that the collective mean of the field is unity.

        See the :ref:`documentation <painting-mesh>` on painting for more
        details on painting catalogs to a mesh.

        Returns
        -------
        real : :class:`pmesh.pm.RealField`
            the painted real field; this has a ``attrs`` dict storing meta-data
        """
        # check for 'Position' column
        if self.position not in self:
            msg = "in order to paint a CatalogSource to a RealField, add a "
            msg += "column named '%s', representing the particle positions" %self.position
            raise ValueError(msg)

        pm = self.pm
        Nlocal = 0 # (unweighted) number of particles read on local rank
        Wlocal = 0 # (weighted) number of particles read on local rank

        # the paint brush window
        paintbrush = window.methods[self.window]

        # initialize the RealField to return
        if out is not None:
            assert isinstance(out, RealField), "output of to_real_field must be a RealField"
            numpy.testing.assert_array_equal(out.pm.Nmesh, pm.Nmesh)
            toret = out
        else:
            toret = RealField(pm)
            toret[:] = 0

        # for interlacing, we need two empty meshes if out was provided
        # since out may have non-zero elements, messing up our interlacing sum
        if self.interlaced:

            real1 = RealField(pm)
            real1[:] = 0

            # the second, shifted mesh (always needed)
            real2 = RealField(pm)
            real2[:] = 0

        # read the necessary data (as dask arrays)
        columns = [self.position, self.weight, self.value, self.selection]

        Position, Weight, Value, Selection = self.read(columns)

        # ensure the slices are synced, since decomposition is collective
        Nlocalmax = max(pm.comm.allgather(len(Position)))

        # paint data in chunks on each rank;
        # we do this by chunk 8 million is pretty big anyways.
        chunksize = _global_options['paint_chunk_size']
        for i in range(0, Nlocalmax, chunksize):
            s = slice(i, i + chunksize)

            if len(Position) != 0:

                # selection has to be computed many times when data is `large`.
                sel = self.base.compute(Selection[s])

                # be sure to use the source to compute
                position, weight, value = \
                    self.base.compute(Position[s], Weight[s], Value[s])

                # FIXME: investigate if move selection before compute
                # speeds up IO.
                position = position[sel]
                weight = weight[sel]
                value = value[sel]
            else:
                # workaround a potential dask issue on empty dask arrays
                position = numpy.empty((0, 3), dtype=Position.dtype)
                weight = None
                value = None
                selection = None

            if weight is None:
                weight = numpy.ones(len(position))

            if value is None:
                value = numpy.ones(len(position))

            # track total (selected) number and sum of weights
            Nlocal += len(position)
            Wlocal += weight.sum()

            # no interlacing
            if not self.interlaced:
                lay = pm.decompose(position, smoothing=0.5 * paintbrush.support)
                p = lay.exchange(position)
                w = lay.exchange(weight)
                v = lay.exchange(value)
                pm.paint(p, mass=w * v, resampler=paintbrush, hold=True, out=toret)

            # interlacing: use 2 meshes separated by 1/2 cell size
            else:
                lay = pm.decompose(position, smoothing=1.0 * paintbrush.support)
                p = lay.exchange(position)
                w = lay.exchange(weight)
                v = lay.exchange(value)

                H = pm.BoxSize / pm.Nmesh

                # in mesh units
                shifted = pm.affine.shift(0.5)

                # paint to two shifted meshes
                pm.paint(p, mass=w * v, resampler=paintbrush, hold=True, out=real1)
                pm.paint(p, mass=w * v, resampler=paintbrush, transform=shifted, hold=True, out=real2)

            Nglobal = pm.comm.allreduce(Nlocal)

            if pm.comm.rank == 0:
                self.logger.info("painted %d out of %d objects to mesh"
                    % (Nglobal, self.base.csize))

        # now the loop over particles is done

        if not self.interlaced:
            # nothing to do, toret is already filled.
            pass
        else:
            # compose the two interlaced fields into the final result.
            c1 = real1.r2c()
            c2 = real2.r2c()

            # and then combine
            for k, s1, s2 in zip(c1.slabs.x, c1.slabs, c2.slabs):
                kH = sum(k[i] * H[i] for i in range(3))
                s1[...] = s1[...] * 0.5 + s2[...] * 0.5 * numpy.exp(0.5 * 1j * kH)

            # FFT back to real-space
            # NOTE: cannot use "toret" here in case user supplied "out"
            c1.c2r(real1)

            # need to add to the returned mesh if user supplied "out"
            toret[:] += real1[:]


        # unweighted number of objects
        N = pm.comm.allreduce(Nlocal)

        # weighted number of objects
        W = pm.comm.allreduce(Wlocal)

        # weighted number density (objs/cell)
        nbar = 1. * W / numpy.prod(pm.Nmesh)

        # make sure we painted something or nbar is nan; in which case
        # we set the density to uniform everywhere.
        if N == 0:
            warnings.warn(("trying to paint particle source to mesh, "
                           "but no particles were found!"),
                            RuntimeWarning
                        )

        # shot noise is volume / un-weighted number
        shotnoise = numpy.prod(pm.BoxSize) / N

        # save some meta-data
        toret.attrs = {}
        toret.attrs['shotnoise'] = shotnoise
        toret.attrs['N'] = N
        toret.attrs['W'] = W
        toret.attrs['num_per_cell'] = nbar

        csum = toret.csum()
        if pm.comm.rank == 0:
            self.logger.info("painted %d out of %d objects to mesh" %(N,self.base.csize))
            self.logger.info("mean particles per cell is %g", nbar)
            self.logger.info("sum is %g ", csum)
            self.logger.info("normalized the convention to 1 + delta")

        if normalize:
            if nbar > 0:
                toret[...] /= nbar
            else:
                toret[...] = 1

        return toret

    @property
    def actions(self):
        """
        The actions to apply to the interpolated density field, optionally
        included the compensation correction.
        """
        actions = MeshSource.actions.fget(self)
        if self.compensated:
            actions = self._get_compensation() + actions
        return actions

    def _get_compensation(self):
        """
        Return the compensation function, which corrects for the
        windowing kernel.

        The compensation function is computed as:

        - if ``interlaced = True``:
          - :func:`CompensateCIC` if using CIC window
          - :func:`CompensateTSC` if using TSC window
          - :func:`CompensatePCS` if using PCS window
        - if ``interlaced = False``:
          - :func:`CompensateCICShotnoise` if using CIC window
          - :func:`CompensateTSCShotnoise` if using TSC window
          - :func:`CompensatePCSShotnoise` if using PCS window
        """
        if self.interlaced:
            d = {'cic' : self.CompensateCIC,
                 'tsc' : self.CompensateTSC,
                 'pcs' : self.CompensatePCS,
                }
        else:
            d = {'cic' : self.CompensateCICShotnoise,
                 'tsc' : self.CompensateTSCShotnoise,
                 'pcs' : self.CompensatePCSShotnoise,
                }

        if not self.window in d:
            raise ValueError("compensation for window %s is not defined" % self.window)

        filter = d[self.window]

        return [('complex', filter, "circular")]

    @staticmethod
    def CompensateTSC(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the TSC window function in configuration space.

        .. note::
            see equation 18 (with p=3) of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            tmp = (numpy.sinc(0.5 * wi / numpy.pi) ) ** 3
            v = v / tmp
        return v

    @staticmethod
    def CompensatePCS(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the PCS window function in configuration space.

        .. note::
            see equation 18 (with p=4) of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            tmp = (numpy.sinc(0.5 * wi / numpy.pi) ) ** 4
            v = v / tmp
        return v

    @staticmethod
    def CompensateCIC(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the CIC window function in configuration space

        .. note::
            see equation 18 (with p=2) of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            tmp = (numpy.sinc(0.5 * wi / numpy.pi) ) ** 2
            tmp[wi == 0.] = 1.
            v = v / tmp
        return v

    @staticmethod
    def CompensateTSCShotnoise(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the TSC window function in configuration space,
        as well as the approximate aliasing correction to the first order

        .. note::
            see equation 20 of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            s = numpy.sin(0.5 * wi)**2
            v = v / (1 - s + 2./15 * s**2) ** 0.5
        return v

    @staticmethod
    def CompensatePCSShotnoise(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the PCS window function in configuration space,
        as well as the approximate aliasing correction to first order

        .. note::

            YF: I derived this by fitting the result to s ** 3
            according to the form given in equation 20 of Jing et al.
            It must be possible to derive this manually.

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            s = numpy.sin(0.5 * wi)**2
            v = v / (1 - 4./3. * s + 2./5. * s**2 - 4./315. * s**3) ** 0.5
        return v

    @staticmethod
    def CompensateCICShotnoise(w, v):
        """
        Return the Fourier-space kernel that accounts for the convolution of
        the gridded field with the CIC window function in configuration space,
        as well as the approximate aliasing correction to the first order

        .. note::
            see equation 20 of
            `Jing et al 2005 <https://arxiv.org/abs/astro-ph/0409240>`_

        Parameters
        ----------
        w : list of arrays
            the list of "circular" coordinate arrays, ranging from
            :math:`[-\pi, \pi)`.
        v : array_like
            the field array
        """
        for i in range(3):
            wi = w[i]
            v = v / (1 - 2. / 3 * numpy.sin(0.5 * wi) ** 2) ** 0.5
        return v

    def save(self, output, dataset='Field', mode='real'):
        """
        Save the mesh as a :class:`~nbodykit.source.mesh.bigfile.BigFileMesh`
        on disk, either in real or complex space.

        Parameters
        ----------
        output : str
            name of the bigfile file
        dataset : str, optional
            name of the bigfile data set where the field is stored
        mode : str, optional
            real or complex; the form of the field to store
        """
        return MeshSource.save(self, output, dataset=dataset, mode=mode)



# Found from https://gist.github.com/remram44/6540454

from six.moves.html_parser import HTMLParser
from six.moves.urllib.error import HTTPError
from six.moves.urllib.request import urlopen
from six import string_types, PY2
import os
import re


# where the main nbodykit data examples live
data_url = "http://portal.nersc.gov/project/m779/nbodykit/example-data"

re_url = re.compile(r'^(([a-zA-Z_-]+)://([^/]+))(/.*)?$')

def resolve_link(link, url):
    m = re_url.match(link)
    if m is not None:
        if not m.group(4):
            # http://domain -> http://domain/
            return link + '/'
        else:
            return link
    elif link[0] == '/':
        # /some/path
        murl = re_url.match(url)
        return murl.group(1) + link
    else:
        # relative/path
        if url[-1] == '/':
            return url + link
        else:
            return url + '/' + link


class ListingParser(HTMLParser):
    """Parses an HTML file and build a list of links.

    Links are stored into the 'links' set. They are resolved into absolute
    links.
    """
    def __init__(self, url):
        HTMLParser.__init__(self)

        if url[-1] != '/':
            url += '/'
        self.__url = url
        self.links = set()

    def handle_starttag(self, tag, attrs):
        if tag == 'a':
            for key, value in attrs:
                if key == 'href':
                    if not value:
                        continue
                    value = resolve_link(value, self.__url)
                    self.links.add(value)
                    break


def mirror(url, target=None):
    """
    Mirror a URL recursively to a local target.

    If ``target`` is not supplied, the last part of the url is used as
    the target.

    Parameters
    ----------
    url : str
        the URL to download
    target : str, optional
        the local file target to save the url to; if not provided, the
        last part of the url is used.
    """
    if target is None:
        target = os.path.normpath(url).split(os.path.sep)[-1]

    def mkdir():
        if not mkdir.done:
            try:
                os.mkdir(target)
            except OSError:
                pass
            mkdir.done = True
    mkdir.done = False

    # open the URL so we can parse it
    response = urlopen(url)

    # HTML file --> keep parsing
    info = response.info()
    content_type = info.type if PY2 else info.get_content_type()
    if content_type == 'text/html':
        contents = response.read().decode()

        parser = ListingParser(url)
        parser.feed(contents)
        for link in parser.links:
            link = resolve_link(link, url)
            if link[-1] == '/':
                link = link[:-1]
            if not link.startswith(url):
                continue
            name = link.rsplit('/', 1)[1]
            if '?' in name:
                continue
            mkdir()
            mirror(link, os.path.join(target, name))
        if not mkdir.done:
            # We didn't find anything to write inside this directory
            # Maybe it's a HTML file?
            if url[-1] != '/':
                end = target[-5:].lower()
                if not (end.endswith('.htm') or end.endswith('.html')):
                    target = target + '.html'
                with open(target, 'wb') as fp:
                    fp.write(contents.encode())
    # just download the file
    else:
        buffer_size = 4096*32
        with open(target, 'wb') as fp:
            chunk = response.read(buffer_size)
            while chunk:
                fp.write(chunk)
                chunk = response.read(buffer_size)

def available_examples():
    """
    Return a list of available example data files from the nbodykit
    data repository on NERSC.

    Returns
    -------
    examples : list
        list of the available file names for download
    """
    # read the contents of the main data URL
    response = urlopen(data_url)
    contents = response.read().decode()

    # parse the available files
    parser = ListingParser(data_url)
    parser.feed(contents)

    # get relative paths and remove bad links
    available = [os.path.relpath(link, data_url) for link in parser.links]
    available = [link for link in available if not any(link.startswith(bad) for bad in ['.', '?'])]
    return sorted(available)


def download_example_data(filenames, download_dirname=None):
    """
    Download a data file from the nbodykit repository of example data.

    For a list of valid file names, see :func:`available_examples`.

    Parameters
    ----------
    filenames : str, list of str
        the name(s) of the example file to download (relative to the path of the
        nbodykit repository); see :func:`available_examples` for the example
        file names
    download_dirname : str, optional
        a local directory to download the file to; if not specified, the
        file will be downloaded to the current working directory
    """
    if isinstance(filenames, string_types):
        filenames = [filenames]

    # make sure the download directory exists
    if download_dirname is not None:
        if not os.path.isdir(download_dirname):
            raise ValueError("specified download directory is not valid")

    # download all requested filenames
    for filename in filenames:

        # where we are saving locally
        if download_dirname is not None:
            target = os.path.join(download_dirname, filename)
        else:
            target = None

        # the full url to the data we want
        url = os.path.join(data_url, filename)

        # try to mirror locally
        try:
            mirror(url, target=target)
        except HTTPError as err:

            # if not found, print available file names, else just raise
            if err.code == 404:
                args = (filename, str(available_examples()))
                raise ValueError("no such example file '%s'\n\navailable examples are: %s" % args)
            else:
                raise



import numpy

class HalofitPower(object):
    """
    Nonlinear power spectrum computed using HaloFit via CLASS.

    Parameters
    ----------
    cosmo : :class:`Cosmology`, astropy.cosmology.FLRW
        the Cosmology instance; astropy cosmology objects are automatically
        converted
    redshift : float
        the redshift of the power spectrum

    Attributes
    ----------
    cosmo : class:`Cosmology`
        the object giving the cosmological parameters
    sigma8 : float
        the z=0 amplitude of matter fluctuations
    redshift : float
        the redshift to compute the power at
    """
    def __init__(self, cosmo, redshift):
        from astropy.cosmology import FLRW

        # convert astropy
        if isinstance(cosmo, FLRW):
            from nbodykit.cosmology import Cosmology
            cosmo = Cosmology.from_astropy(cosmo)

        # internal cosmology clone with nonlinear enabled
        self.cosmo = cosmo.clone(nonlinear=True)
        self.redshift = redshift
        self._sigma8 = self.cosmo.sigma8

        # store meta-data
        self._attrs = {}
        self._attrs['cosmo'] = dict(cosmo)

    @property
    def attrs(self):
        """
        The meta-data dictionary
        """
        self._attrs['redshift'] = self.redshift
        self._attrs['sigma8'] = self.sigma8
        return self._attrs

    @property
    def sigma8(self):
        """
        The amplitude of matter fluctuations at :math:`z=0`.
        """
        return self._sigma8

    @sigma8.setter
    def sigma8(self, value):
        self._sigma8 = value
        self.cosmo = self.cosmo.match(sigma8=value)
        self._attrs['cosmo'] = dict(self.cosmo)

    def __call__(self, k):
        r"""
        Return the power in units of :math:`h^{-3} \mathrm{Mpc}^3`.

        Parameters
        ----------
        k : float, array_like
            the wavenumbers in units of :math:`h \mathrm{Mpc}^{-1}`

        Returns
        -------
        Pk : float, array_like
            the linear power spectrum evaluated at ``k`` in units of
            :math:`h^{-3} \mathrm{Mpc}^3`
        """
        k = numpy.asarray(k)
        if k.max() > self.cosmo.P_k_max:
            msg = "results can only be computed up to k=%.2e h/Mpc; " %self.cosmo.P_k_max
            msg += "try increasing the Cosmology parameter 'P_k_max'"
            raise ValueError(msg)

        kmin = self.cosmo.P_k_min
        inrange = k > 1.00001*kmin

        Pk = numpy.zeros_like(k)
        k_in = k[inrange]; k_out = k[~inrange]

        # nonlinear power in range
        Pk[inrange] = self.cosmo.get_pk(k=k_in, z=self.redshift)

        # linear power on large scales (small k)
        Pk[~inrange] = self.cosmo.get_pklin(k=k_out, z=self.redshift)
        return Pk



from __future__ import absolute_import
# the future import is important. or in python 2.7 we try to
# import this module itself. Due to the unfortnate name conflict!

import numpy

from .base import FileType
from six import string_types
import json
from nbodykit.utils import JSONDecoder

class Automatic: pass

class BigFile(FileType):
    """
    A file object to handle the reading of columns of data from
    a :mod:`bigfile` file.

    :mod:`bigfile` is a reproducible, massively parallel IO library for
    large, hierarchical datasets, and it is the default format of the
    `FastPM <https://github.com/rainwoodman/fastpm>`_ and the
    `MP-Gadget <https://github.com/bluetides-project/MP-Gadget>`_
    simulations.

    See also: https://github.com/rainwoodman/bigfile

    Parameters
    ----------
    path : str
        the name of the directory holding the bigfile data
    exclude : list of str, optional
        the data sets to exlude from loading within bigfile; default
        is the header
    header : str, optional
        the path to the header; default is to use a column 'Header'.
        It is relative to the file, not the dataset.
    dataset : str
        load a specific dataset from the bigfile; default is to starting
        from the root.
    """
    def __init__(self, path, exclude=None, header=Automatic, dataset='./'):

        if not dataset.endswith('/'): dataset = dataset + '/'

        import bigfile

        self.dataset = dataset
        self.path = path

        # store the attributes
        self.attrs = {}

        # the file path
        with bigfile.BigFile(filename=path) as ff:
            columns = ff[self.dataset].blocks
            if header is Automatic:
                for header in ['Header', 'header', './']:
                    if header in columns: break

            if exclude is None:
                exclude = [header]

            columns = list(set(columns) - set(exclude))

            ds = bigfile.BigData(ff[self.dataset], columns)

            # set the data type and size
            self.dtype = ds.dtype
            self.size = ds.size

            header = ff[header]
            attrs = header.attrs

            # copy over the attrs
            for k in attrs.keys():

                # load a JSON representation if str starts with json:://
                if isinstance(attrs[k], string_types) and attrs[k].startswith('json://'):
                    self.attrs[k] = json.loads(attrs[k][7:], cls=JSONDecoder)
                # copy over an array
                else:
                    self.attrs[k] = numpy.array(attrs[k], copy=True)

    def read(self, columns, start, stop, step=1):
        """
        Read the specified column(s) over the given range,
        as a dictionary

        'start' and 'stop' should be between 0 and :attr:`size`,
        which is the total size of the binary file (in particles)
        """
        import bigfile
        if isinstance(columns, string_types): columns = [columns]

        with bigfile.BigFile(filename=self.path)[self.dataset] as f:
            ds = bigfile.BigData(f, columns)
            return ds[start:stop][::step]



__all__ = ['notebook']

import matplotlib
import os
_cwd = os.path.split(os.path.abspath(__file__))[0]

for name in __all__:
    path = os.path.join(_cwd, 'notebook.mplstyle')
    globals()[name] = matplotlib.rc_params_from_file(path, use_default_template=False)



import numpy
import logging
from pmesh.pm import ParticleMesh, RealField, ComplexField

class MeshSource(object):
    """
    Base class for a source in the form of data on an input grid.

    The MeshSource object remembers the original source together with
    a sequence of transformations (added via the apply method).

    ``dtype`` is the type of the real numbers, either 'f4' or 'f8'.

    .. note:

        Subclasses of this class must implement either :func:`to_real_field`
        or :func:`to_complex_field`. The methods must return either a
        :class:`pmesh.pm.RealField` or a :class:`pmesh.pm.ComplexField` object.

    Parameters
    ----------
    comm :
        the global MPI communicator
    Nmesh : int, array_like
        the number of cells per grid size
    BoxSize : array_like
        the size of the box
    dtype : str
        the desired data type of the grid
    """
    logger = logging.getLogger('MeshSource')

    def __init__(self, comm, Nmesh, BoxSize, dtype):

        # ensure self.comm is set, though usually already set by the child.
        self.comm = comm
        self.dtype = dtype

        if Nmesh is None or BoxSize is None:
            raise ValueError("both Nmesh and BoxSize must not be None to initialize ParticleMesh")

        Nmesh = numpy.array(Nmesh)
        if Nmesh.ndim == 0:
            ndim = 3
        else:
            ndim = len(Nmesh)
        _Nmesh = numpy.empty(ndim, dtype='i8')
        _Nmesh[:] = Nmesh
        self.pm = ParticleMesh(BoxSize=BoxSize, Nmesh=_Nmesh,
                                dtype=self.dtype, comm=self.comm)

        self.attrs['BoxSize'] = self.pm.BoxSize.copy()
        self.attrs['Nmesh'] = self.pm.Nmesh.copy()

        self._actions = []
        self.base = None

    def __finalize__(self, other):
        """
        Finalize the creation of a MeshSource object by copying over
        attributes from a second MeshSource.

        Parameters
        ----------
        other : MeshSource
            the second MeshSource to copy over attributes from
        """
        if isinstance(other, MeshSource):
            self.comm = other.comm
            self.dtype = other.dtype
            self.pm = other.pm
            self.attrs.update(other.attrs)
            self._actions = []
            self.actions.extend(other.actions)

        return self


    def view(self):
        """
        Return a "view" of the MeshSource, in the spirit of
        numpy's ndarray view.

        This returns a new MeshSource whose memory is owned by ``self``.

        Note that for CatalogMesh objects, this is overidden by the
        ``CatalogSource.view`` function.
        """
        view = object.__new__(MeshSource)
        view.base = self
        return view.__finalize__(self)

    @property
    def attrs(self):
        """
        A dictionary storing relevant meta-data about the CatalogSource.
        """
        try:
            return self._attrs
        except AttributeError:
            self._attrs = {}
            return self._attrs

    @property
    def actions(self):
        """
        A list of actions to apply to the density field when interpolating
        to the mesh.

        This stores tuples of ``(mode, func, kind)``; see :func:`apply`
        for more details.
        """
        return self._actions

    def apply(self, func, kind='wavenumber', mode='complex'):
        """
        Return a view of the mesh, with :attr:`actions` updated to
        apply the specified function, either in Fourier space or
        configuration space, based on ``mode``

        Parameters
        ----------
        func : callable
            func(x, y) where x is a list of ``r`` (``k``) values that broadcasts
            into a full array, when ``mode`` is 'real' ('complex');
            the value of x depends on ``kind``. ``y`` is the value of
            the mesh field on the corresponding locations.
        kind : string, optional
            The kind of value in x.

            - When ``mode`` is 'complex':

              - 'wavenumber' means wavenumber from [- 2 pi / L * N / 2, 2 pi / L * N / 2).
              - 'circular' means circular frequency from [- pi, pi).
              - 'index' means [0, Nmesh )

            - When ``mode`` is 'real':

              - 'relative' means distance from [-0.5 Boxsize, 0.5 BoxSize).
              - 'index' means [0, Nmesh )
        mode : 'complex' or 'real', optional
            whether to apply the function to the mesh in configuration space
            or Fourier space

        Returns
        -------
        MeshSource :
            a view of the mesh object with the :attr:`actions` attribute
            updated to include the new action
        """
        assert mode in ['complex', 'real'], "``mode`` should be 'complex' or 'real'"
        if mode == 'real':
            assert kind in ['relative', 'index']
        else:
            assert kind in ['wavenumber', 'circular', 'index']

        view = self.view()
        view.actions.append((mode, func, kind))
        return view

    def __len__(self):
        """
        Length of a mesh source is zero
        """
        return 0

    def to_real_field(self, out=None, normalize=True):
        """
        Convert the mesh source to the configuration-space field,
        returning a :class:`pmesh.pm.RealField` object.

        Not implemented in the base class, unless object is a view.
        """
        if isinstance(self.base, MeshSource): return self.base.to_real_field()
        return NotImplemented

    def to_complex_field(self, out=None):
        """
        Convert the mesh source to the Fourier-space field,
        returning a :class:`pmesh.pm.ComplexField` object.

        Not implemented in the base class, unless object is a view.
        """
        if isinstance(self.base, MeshSource): return self.base.to_complex_field()
        return NotImplemented

    def to_field(self, mode='real', out=None):
        """
        Return the mesh as a :mod:`pmesh` Field object, either in Fourier
        space or configuration space, based on ``mode``.

        This will call :func:`to_real_field` or :func:`to_complex_field`
        based on ``mode``.

        Parameters
        ----------
        mode : 'real' or 'complex'
            the return type of the field

        Returns
        -------
        :class:`~pmesh.pm.RealField`, :class:`~pmesh.pm.ComplexField` :
            either a RealField of ComplexField, storing the value of the field
            on the mesh
        """
        if mode == 'real':
            real = self.to_real_field()
            if real is NotImplemented:
                complex = self.to_complex_field()
                assert complex is not NotImplemented
                real = complex.c2r(out=Ellipsis)
                if hasattr(complex, 'attrs'):
                    real.attrs = complex.attrs
            var = real
        elif mode == 'complex':
            complex = self.to_complex_field()
            if complex is NotImplemented:
                real = self.to_real_field()
                assert real is not NotImplemented
                complex = real.r2c(out=Ellipsis)
                if hasattr(real, 'attrs'):
                    complex.attrs = real.attrs
            var = complex
        else:
            raise ValueError("mode is either real or complex, %s given" % mode)

        return var

    def paint(self, mode="real", Nmesh=None):
        """
        Paint the density on the mesh and apply
        any transformation functions specified in :attr:`actions`.

        The return type of the :mod:`pmesh` Field object is specified by
        ``mode``. This calls :func:`to_field` to convert the mesh to a Field.

        See the :ref:`documentation <painting-mesh>` on painting for more
        details on painting catalogs to a mesh.

        Parameters
        ----------
        mode : 'real' or 'complex'
            the type of the returned Field object, either a RealField or
            ComplexField
        Nmesh : int or array_like, or None
            If given and different from the intrinsic Nmesh of the source,
            resample the mesh to the given resolution

        Returns
        -------
        :class:`~pmesh.pm.RealField`, :class:`~pmesh.pm.ComplexField` :
            either a RealField of ComplexField, with the functions in
            :attr:`actions` applied to it
        """
        if not mode in ['real', 'complex']:
            raise ValueError('mode must be "real" or "complex"')

        # add a dummy action to ensure the right mode of return value
        actions = self.actions + [(mode, )]

        # if we expect complex, be smart and use complex directly.
        var = self.to_field(mode=actions[0][0])

        if not hasattr(var, 'attrs'):
            attrs = {}
        else:
            attrs = var.attrs

        for action in actions:
            # ensure var is the right mode

            if action[0] == 'complex':
                if not isinstance(var, ComplexField):
                    var = var.r2c(out=Ellipsis)
            if action[0] == 'real':
                if not isinstance(var, RealField):
                    var = var.c2r(out=Ellipsis)

            if len(action) > 1:
                # there is a filter function
                kwargs = {}
                kwargs['func'] = action[1]
                if action[2] is not None:
                    kwargs['kind'] = action[2]
                kwargs['out'] = Ellipsis
                var.apply(**kwargs)

        pm = self.pm.resize(Nmesh)

        if any(pm.Nmesh != self.pm.Nmesh):
            # resample if the output mesh mismatches
            # XXX: this could be done more efficiently.
            var1 = pm.create(mode=mode)
            var.resample(out=var1)
            var = var1

            if self.comm.rank == 0:
                self.logger.info('%s resampling from %s to %s done' % (str(self), str(self.pm.Nmesh), str(pm.Nmesh)))

        var.attrs = attrs
        var.attrs.update(self.attrs)

        if self.comm.rank == 0:
            self.logger.info('field: %s painting done' % str(self))

        return var

    def preview(self, axes=None, Nmesh=None, root=0):
        """
        Gather the mesh into as a numpy array, with
        (reduced) resolution. The result is broadcast to
        all ranks, so this uses :math:`\mathrm{Nmesh}^3` per rank.

        Parameters
        ----------
        Nmesh : int, array_like
            The desired Nmesh of the result. Be aware this function
            allocates memory to hold a full ``Nmesh`` on each rank.
        axes : int, array_like
            The axes to project the preview onto., e.g. (0, 1)
        root : int, optional
            the rank number to treat as root when gathering to a single rank

        Returns
        -------
        out : array_like
            An numpy array holding the real density field.
        """
        field = self.to_field(mode='real')
        if Nmesh is None:
            Nmesh = self.pm.Nmesh

        return field.preview(Nmesh, axes=axes)

    def save(self, output, dataset='Field', mode='real'):
        """
        Save the mesh as a :class:`~nbodykit.source.mesh.bigfile.BigFileMesh`
        on disk, either in real or complex space.

        Parameters
        ----------
        output : str
            name of the bigfile file
        dataset : str, optional
            name of the bigfile data set where the field is stored
        mode : str, optional
            real or complex; the form of the field to store
        """
        import bigfile
        import warnings
        import json
        from nbodykit.utils import JSONEncoder

        field = self.paint(mode=mode)

        with bigfile.BigFileMPI(self.pm.comm, output, create=True) as ff:
            data = numpy.empty(shape=field.size, dtype=field.dtype)
            field.ravel(out=data)
            with ff.create_from_array(dataset, data) as bb:
                if isinstance(field, RealField):
                    bb.attrs['ndarray.shape'] = field.pm.Nmesh
                    bb.attrs['BoxSize'] = field.pm.BoxSize
                    bb.attrs['Nmesh'] = field.pm.Nmesh
                elif isinstance(field, ComplexField):
                    bb.attrs['ndarray.shape'] = field.Nmesh, field.Nmesh, field.Nmesh // 2 + 1
                    bb.attrs['BoxSize'] = field.pm.BoxSize
                    bb.attrs['Nmesh'] = field.pm.Nmesh

                for key in field.attrs:
                    # do not override the above values -- they are vectors (from pm)
                    if key in bb.attrs: continue
                    value = field.attrs[key]
                    try:
                        bb.attrs[key] = value
                    except ValueError:
                        try:
                            json_str = 'json://'+json.dumps(value, cls=JSONEncoder)
                            bb.attrs[key] = json_str
                        except:
                            warnings.warn("attribute %s of type %s is unsupported and lost while saving MeshSource" % (key, type(value)))



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

import os
import pytest
import numpy
from numpy.testing import assert_array_equal, assert_allclose
from kdcount import correlate, sphere

# debug logging
setup_logging()

def gather_data(source, name):
    return numpy.concatenate(source.comm.allgather(source[name].compute()), axis=0)

def generate_survey_data(seed):
    s = RandomCatalog(1000, seed=seed)
    s['RA'] = s.rng.uniform(low=50, high=260, size=s.size)
    s['DEC'] = s.rng.uniform(low=-10.6, high=60., size=s.size)
    return s

def generate_sim_data(seed):
    s = UniformCatalog(nbar=1000, BoxSize=1.0, seed=seed)
    s['RA'], s['DEC'] = transform.CartesianToEquatorial(s['Position'], observer=0.5*s.attrs['BoxSize'])
    return s

def reference_paircount(pos1, w1, edges, pos2=None, w2=None):
    """Reference pair counting via kdcount"""
    # set up the trees
    tree1 = sphere.points(*pos1, boxsize=None, weights=w1)
    if pos2 is None:
        tree2 = tree1
    else:
        tree2 = sphere.points(*pos2, boxsize=None, weights=w2)

    # run the pair count
    bins = sphere.AngularBinning(edges)
    pc = correlate.paircount(tree1, tree2, bins, np=0, usefast=False, compute_mean_coords=True)

    return numpy.nan_to_num(pc.pair_counts), numpy.nan_to_num(pc.mean_centers), pc.sum1

@MPITest([1, 4])
def test_survey_auto(comm):
    CurrentMPIComm.set(comm)

    # random particles
    source = generate_survey_data(seed=42)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform(size=len(source))

    # make the bin edges
    edges = numpy.linspace(0.001, 1.0, 10)

    # do the weighted paircount
    r = SurveyDataPairCount('angular', source, edges, weight='Weight', show_progress=False)

    # cannot compute theta=0
    with pytest.raises(ValueError):
        r = SurveyDataPairCount('angular', source, numpy.linspace(0, 1.0, 10))

    ra = gather_data(source, 'RA')
    dec = gather_data(source, 'DEC')
    w = gather_data(source, 'Weight')

    # verify with kdcount
    npairs, thetaavg, wsum = reference_paircount([ra,dec], w, edges)
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['npairs'] * r.pairs['weightavg'])

@MPITest([1, 4])
def test_survey_cross(comm):
    CurrentMPIComm.set(comm)

    # random particles with weights
    first = generate_survey_data(seed=42)
    first['Weight'] = first.rng.uniform(size=first.size)
    second = generate_survey_data(seed=84)
    second['Weight'] = second.rng.uniform(size=second.size)

    # make the bin edges
    edges = numpy.linspace(0.001, 1.0, 10)

    # do the paircount
    r = SurveyDataPairCount('angular', first, edges, second=second)

    # gather data to run kdcount
    ra1 = gather_data(first, 'RA')
    ra2 = gather_data(second, 'RA')
    dec1 = gather_data(first, 'DEC')
    dec2 = gather_data(second, 'DEC')
    w1 = gather_data(first, 'Weight')
    w2 = gather_data(second, 'Weight')

    # verify with kdcount
    npairs, thetaavg, wsum = reference_paircount([ra1,dec1], w1, edges, pos2=[ra2,dec2], w2=w2)
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['npairs'] * r.pairs['weightavg'])

    # test save
    r.save('angular-paircount-test.json')
    r2 = SurveyDataPairCount.load('angular-paircount-test.json')
    assert_array_equal(r.pairs.data, r2.pairs.data)

    if comm.rank == 0: os.remove('angular-paircount-test.json')

@MPITest([1, 4])
def test_sim_auto(comm):
    CurrentMPIComm.set(comm)

    # random particles
    source = generate_sim_data(seed=42)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform(size=len(source))

    # make the bin edges
    edges = numpy.linspace(0.001, 1.0, 10)

    # do the weighted paircount
    r = SimulationBoxPairCount('angular', source, edges, weight='Weight', show_progress=False)

    # cannot compute theta=0
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('angular', source, numpy.linspace(0, 1.0, 10))

    ra = gather_data(source, 'RA')
    dec = gather_data(source, 'DEC')
    w = gather_data(source, 'Weight')

    # verify with kdcount
    npairs, thetaavg, wsum = reference_paircount([ra,dec], w, edges)
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['npairs'] * r.pairs['weightavg'])

@MPITest([1, 4])
def test_sim_cross(comm):
    CurrentMPIComm.set(comm)

    # random particles with weights
    first = generate_sim_data(seed=42)
    first['Weight'] = first.rng.uniform(size=first.size)
    second = generate_sim_data(seed=84)
    second['Weight'] = second.rng.uniform(size=second.size)

    # make the bin edges
    edges = numpy.linspace(0.001, 1.0, 10)

    # do the paircount
    r = SimulationBoxPairCount('angular', first, edges, second=second)

    # gather data to run kdcount
    ra1, ra2 = gather_data(first, 'RA'), gather_data(second, 'RA')
    dec1, dec2 = gather_data(first, 'DEC'), gather_data(second, 'DEC')
    w1, w2 = gather_data(first, 'Weight'), gather_data(second, 'Weight')

    # verify with kdcount
    npairs, thetaavg, wsum = reference_paircount([ra1,dec1], w1, edges, pos2=[ra2,dec2], w2=w2)
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wsum, r.pairs['npairs'] * r.pairs['weightavg'])

    # test save
    r.save('angular-paircount-test.json')
    r2 = SimulationBoxPairCount.load('angular-paircount-test.json')
    assert_array_equal(r.pairs.data, r2.pairs.data)

    if comm.rank == 0: os.remove('angular-paircount-test.json')


@MPITest([1])
def test_missing_columns(comm):

    CurrentMPIComm.set(comm)
    source = generate_survey_data(seed=42)
    edges = numpy.linspace(0.01, 10.0, 10)

    # missing column
    with pytest.raises(ValueError):
        r = SurveyDataPairCount('angular', source, edges, ra='BAD')



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit.tutorials import DemoHaloCatalog
from nbodykit import setup_logging
import shutil
import pytest


setup_logging()

@MPITest([4])
def test_no_seed(comm):

    CurrentMPIComm.set(comm)

    halos = DemoHaloCatalog('bolshoi', 'rockstar', 0.5)
    hod = halos.populate(Zheng07Model)

    # seed is set randomly
    assert hod.attrs['seed'] is not None

@MPITest([4])
def test_bad_model(comm):

    CurrentMPIComm.set(comm)

    halos = DemoHaloCatalog('bolshoi', 'rockstar', 0.5)
    with pytest.raises(TypeError):
        hod = halos.populate('Zheng07Model')


@MPITest([4])
def test_failed_populate(comm):

    CurrentMPIComm.set(comm)

    # the demo halos
    halos = DemoHaloCatalog('bolshoi', 'rockstar', 0.5)

    # initialize model with bad MASS column
    model = Zheng07Model.to_halotools(halos.cosmo, halos.attrs['redshift'], '200c')

    with pytest.raises(Exception):
        hod = halos.populate(model)


@MPITest([4])
def test_no_galaxies(comm):

    CurrentMPIComm.set(comm)
    halos = DemoHaloCatalog('bolshoi', 'rockstar', 0.5)

    # no galaxies populated into halos
    # NOTE: logMmin is unrealistically large here
    with pytest.raises(ValueError):
        hod = halos.populate(Zheng07Model, seed=42, logMmin=17)

@MPITest([4])
def test_repopulate(comm):

    CurrentMPIComm.set(comm)

    # initialize the halos
    halos = DemoHaloCatalog('bolshoi', 'rockstar', 0.5)

    # populate the mock first
    hod = halos.populate(Zheng07Model, seed=42)
    size = hod.csize

    # repopulate (with same seed --> same catalog)
    hod.repopulate(seed=42)
    assert hod.csize == size

    # repopulate with random seed
    # make sure root sets a random seed and it's saved
    hod.repopulate(seed=None)
    assert hod.attrs['seed'] is not None

    # new params, same seed
    hod.repopulate(seed=42, alpha=1.0)
    assert hod.csize != size

    # bad param name
    with pytest.raises(ValueError):
        hod.repopulate(seed=42, bad_param_name=1.0)


@MPITest([4])
def test_hod_cm(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    # lognormal particles
    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=BoxSize, Nmesh=128, seed=42)

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir', posdef='cm')

    # make the HOD catalog from halotools catalog
    hod = halos.populate(Zheng07Model, seed=42)

    # RSD offset in 'z' direction
    hod['Position'] += hod['VelocityOffset'] * [0, 0, 1]

    # compute the power
    r = FFTPower(hod.to_mesh(Nmesh=128), mode='2d', Nmu=5, los=[0,0,1])

@MPITest([4])
def test_hod_peak(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    # lognormal particles
    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=BoxSize, Nmesh=128, seed=42)

    source['Density'] = KDDensity(source).density

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir', posdef='peak')

    # make the HOD catalog from halotools catalog
    hod = halos.populate(Zheng07Model, seed=42)

    # RSD offset in 'z' direction
    hod['Position'] += hod['VelocityOffset'] * [0, 0, 1]

    # compute the power
    r = FFTPower(hod.to_mesh(Nmesh=128), mode='2d', Nmu=5, los=[0,0,1])

@MPITest([4])
def test_save(comm):

    CurrentMPIComm.set(comm)

    redshift = 0.55
    cosmo = cosmology.Planck15
    BoxSize = 512

    # lognormal particles
    Plin = cosmology.LinearPower(cosmo, redshift=redshift, transfer='EisensteinHu')
    source = LogNormalCatalog(Plin=Plin, nbar=3e-3, BoxSize=BoxSize, Nmesh=128, seed=42)

    # run FOF
    r = FOF(source, linking_length=0.2, nmin=20)
    halos = r.to_halos(cosmo=cosmo, redshift=redshift, particle_mass=1e12, mdef='vir')

    # make the HOD catalog from halotools catalog
    hod = halos.populate(Zheng07Model, seed=42)

    # save to a tmp file
    hod.save('tmp-hod.bigfile', ['Position', 'Velocity', 'gal_type'])

    # read tmp file
    cat = BigFileCatalog('tmp-hod.bigfile', header="Header")

    try:
        # check attrs
        for name in hod.attrs:
            numpy.testing.assert_array_equal(cat.attrs[name], hod.attrs[name])

        # check same size
        assert hod.csize == cat.csize

        # check total number of satellites
        nsat1 = comm.allreduce(hod['gal_type'].sum())
        nsat2 = comm.allreduce(cat['gal_type'].sum())
        assert nsat1 == nsat2

    finally:
        comm.barrier()
        if comm.rank == 0:
            shutil.rmtree('tmp-hod.bigfile')



from runtests.mpi import MPITest
from nbodykit.io.binary import BinaryFile
import os
import numpy
import tempfile
import pickle
import pytest

@MPITest([1])
def test_data(comm):

    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'wb') as ff:    
        
        # generate data
        pos = numpy.random.random(size=(1024, 3))
        vel = numpy.random.random(size=(1024, 3))
        pos.tofile(ff); vel.tofile(ff); ff.seek(0)
        
        # read
        f = BinaryFile(ff.name, [('Position', ('f8', 3)), ('Velocity', ('f8', 3))], size=1024)
        
        assert f.size == len(f[:]), "mismatch between 'size' and length of data read"
        assert f.size == 1024, "wrong data size"
        
        numpy.testing.assert_almost_equal(pos, f['Position'][:])
        numpy.testing.assert_almost_equal(vel, f['Velocity'][:])

    os.remove(tmpfile)

@MPITest([1])
def test_offsets(comm):

    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'wb') as ff: 

        # generate data
        pos = numpy.random.random(size=(1024, 3))
        vel = numpy.random.random(size=(1024, 3))
        pos.tofile(ff); vel.tofile(ff); ff.seek(0)

        # pass an offsets dictionary
        dtype = [('Position', ('f8', 3)), ('Velocity', ('f8', 3))]
        f = BinaryFile(ff.name, dtype, size=1024, offsets={'Position':0, 'Velocity':pos.nbytes})

        assert f.size == len(f[:]), "mismatch between 'size' and length of data read"
        assert f.size == 1024, "wrong data size"

        numpy.testing.assert_almost_equal(pos, f['Position'][:])
        numpy.testing.assert_almost_equal(vel, f['Velocity'][:])

        # wrong offsets dict
        with pytest.raises(ValueError):
            f = BinaryFile(ff.name, dtype, size=1024, offsets={'Position':0})

        # must be a dict
        with pytest.raises(TypeError):
            f = BinaryFile(ff.name, dtype, size=1024, offsets=[('Position',0)])
        
    # cleanup
    os.remove(tmpfile)

@MPITest([1])
def test_header(comm):

    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'wb') as ff: 

        # generate data
        pos = numpy.random.random(size=(1024, 3))
        vel = numpy.random.random(size=(1024, 3))
        hdr = numpy.arange(10); hdr.tofile(ff)
        pos.tofile(ff); vel.tofile(ff); ff.seek(0)

        # read
        dtype = [('Position', ('f8', 3)), ('Velocity', ('f8', 3))]
        f = BinaryFile(ff.name, dtype, size=1024, header_size=hdr.nbytes)

        assert f.size == len(f[:]), "error when header is non-zero"
        assert f.size == 1024,"error when header is non-zero"

        # test wrong header size
        with pytest.raises(ValueError):
            f = BinaryFile(ff.name, dtype, header_size=hdr.nbytes-1)
            
    # cleanup
    os.remove(tmpfile)

@MPITest([1])
def test_infer_size(comm):

    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'wb') as ff: 

        # generate data
        pos = numpy.random.random(size=(1024, 3))
        vel = numpy.random.random(size=(1024, 3))
        hdr = numpy.arange(10); hdr.tofile(ff)
        pos.tofile(ff); vel.tofile(ff); ff.seek(0)

        # read
        dtype = [('Position', ('f8', 3)), ('Velocity', ('f8', 3))]
        f = BinaryFile(ff.name, dtype, header_size=hdr.nbytes)

        assert f.size == len(f[:]), "error when trying to infer size"
        assert f.size == 1024, "error when trying to infer size"
    
    os.remove(tmpfile)

@MPITest([1])
def test_wrong_size(comm):

    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'wb') as ff: 

        # generate data
        pos = numpy.random.random(size=(1024, 3))
        vel = numpy.random.random(size=(1024, 3))
        hdr = numpy.arange(10); hdr.tofile(ff)
        pos.tofile(ff); vel.tofile(ff); ff.seek(0)

        # size must be an int
        dtype = [('Position', ('f8', 3)), ('Velocity', ('f8', 3))]
        with pytest.raises(TypeError):
            f = BinaryFile(ff.name, dtype, header_size=hdr.nbytes, size=1024.0)
    
    # cleanup
    os.remove(tmpfile)


@MPITest([1])
def test_pickle(comm):

    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'wb') as ff: 

        # generate data
        pos = numpy.random.random(size=(1024, 3))
        vel = numpy.random.random(size=(1024, 3))
        pos.tofile(ff); vel.tofile(ff); ff.seek(0)

        # read
        f = BinaryFile(ff.name, [('Position', ('f8', 3)), ('Velocity', ('f8', 3))], size=1024)
        
        # pickle and assert equality
        s = pickle.dumps(f)
        f2 = pickle.loads(s)
        numpy.testing.assert_almost_equal(f['Position'][:], f2['Position'][:])
    
    # cleanup
    os.remove(tmpfile)


from nbodykit import GlobalCache
from nbodykit.lab import UniformCatalog

def test_cache():

    cat = UniformCatalog(nbar=10000, BoxSize=1.0)
    cat['test'] = cat['Position'] ** 5
    test = cat['test'].compute()

    # cache should no longer be empty
    cache = GlobalCache.get()
    assert cache.cache.total_bytes > 0

    # resize
    GlobalCache.resize(100)
    assert cache.cache.total_bytes < 100



from nbodykit.transform import ConstantArray
from nbodykit import _global_options, CurrentMPIComm, GlobalCache

from six import string_types, add_metaclass
import numpy
import logging
import warnings
import abc
import inspect
import dask.array as da

class ColumnAccessor(da.Array):
    """
    Provides access to a Column from a Catalog.

    This is a thin subclass of :class:`dask.array.Array` to
    provide a reference to the catalog object,
    an additional ``attrs`` attribute (for recording the
    reproducible meta-data), and some pretty print support.

    Due to particularity of :mod:`dask`, any transformation
    that is not explicitly in-place will return
    a :class:`dask.array.Array`, and losing the pointer to
    the original catalog and the meta data attrs.

    Parameters
    ----------
    catalog : CatalogSource
        the catalog from which the column was accessed
    daskarray : dask.array.Array
        the column in dask array form
    is_default : bool, optional
        whether this column is a default column; default columns are not
        serialized to disk, as they are automatically available as columns
    """
    def __new__(cls, catalog, daskarray, is_default=False):
        self = da.Array.__new__(ColumnAccessor,
                daskarray.dask,
                daskarray.name,
                daskarray.chunks,
                daskarray.dtype,
                daskarray.shape)
        self.catalog = catalog
        self.is_default = is_default
        self.attrs = {}
        return self

    def __getitem__(self, key):

        # compute dask index b/c they are not fully supported
        if isinstance(key, da.Array):
            key = self.catalog.compute(key)

        # base class behavior
        d = da.Array.__getitem__(self, key)

        # return a ColumnAccessor (okay b/c __setitem__ checks for circular references)
        toret = ColumnAccessor(self.catalog, d)
        toret.attrs.update(self.attrs)
        return toret

    def as_daskarray(self):
        return da.Array(
                self.dask,
                self.name,
                self.chunks,
                self.dtype,
                self.shape)

    def compute(self):
        return self.catalog.compute(self)

    def __str__(self):
        r = da.Array.__str__(self)
        if len(self) > 0:
            r = r + " first: %s" % str(self[0].compute())
        if len(self) > 1:
            r = r + " last: %s" % str(self[-1].compute())
        return r

def column(name=None, is_default=False):
    """
    Decorator that defines the decorated function as a column in a
    CatalogSource.

    This can be used as a decorator with or without arguments. If no ``name``
    is specified, the function name is used.

    Parameters
    ----------
    name : str, optional
        the name of the column; if not provided, the name of the function
        being decorated is used
    is_default : bool, optional
        whether the column is a default column; default columns are not
        serialized to disk
    """
    def decorator(getter, name=name):
        getter.column_name = getter.__name__ if name is None else name
        getter.is_default = is_default
        return getter

    # handle the case when decorator was called without arguments
    # in that case "name" is actually the function we are decorating
    if hasattr(name, '__call__'):
        getter = name
        return decorator(getter, name=getter.__name__)
    else:
        return decorator

class ColumnFinder(abc.ABCMeta):
    """
    A meta-class that will register all columns of a class that have
    been marked with the :func:`column` decorator.

    This adds the following attributes to the class definition:

    1. ``_defaults`` : default columns, specified by passing ``default=True`` to
    the :func:`column` decorator.
    2. ``_hardcolumns`` : non-default, hard-coded columns

    .. note::
        This is a subclass of :class:`abc.ABCMeta` so subclasses can
        define abstract properties, if they need to.
    """
    def __init__(cls, clsname, bases, attrs):

        # attach the registry attributes
        cls._defaults = set()
        cls._hardcolumns = set()

        # loop over class and its bases
        classes = inspect.getmro(cls)
        for c in reversed(classes):

            # loop over each attribute
            for name in c.__dict__:
                value = c.__dict__[name]

                # if it's member function implementing a column,
                # record it and check if its a default
                if getattr(value, 'column_name', None):
                    if value.is_default:
                        cls._defaults.add(value.column_name)
                    else:
                        cls._hardcolumns.add(value.column_name)

@add_metaclass(ColumnFinder)
class CatalogSourceBase(object):
    """
    An abstract base class that implements most of the functionality in
    :class:`CatalogSource`.

    The main difference between this class and :class:`CatalogSource` is that
    this base class does not assume the object has a :attr:`size` attribute.

    .. note::
        See the docstring for :class:`CatalogSource`. Most often, users should
        implement custom sources as subclasses of :class:`CatalogSource`.

    The names of hard-coded columns, i.e., those defined through member
    functions of the class, are stored in the :attr:`_defaults` and
    :attr:`_hardcolumns` attributes. These attributes are computed by the
    :class:`ColumnFinder` meta-class.

    Parameters
    ----------
    comm :
        the MPI communicator to use for this object
    """
    logger = logging.getLogger('CatalogSourceBase')

    @staticmethod
    def make_column(array):
        """
        Utility function to convert an array-like object to a
        :class:`dask.array.Array`.

        .. note::
            The dask array chunk size is controlled via the ``dask_chunk_size``
            global option. See :class:`~nbodykit.set_options`.

        Parameters
        ----------
        array : array_like
            an array-like object; can be a dask array, numpy array,
            ColumnAccessor, or other non-scalar array-like object

        Returns
        -------
        :class:`dask.array.Array` :
            a dask array initialized from ``array``
        """
        if isinstance(array, da.Array):
            return array
        elif isinstance(array, ColumnAccessor):
            # important to get the accessor as a dask array to avoid circular
            # references
            return array.as_daskarray()
        else:
            return da.from_array(array, chunks=_global_options['dask_chunk_size'])

    def __new__(cls, *args, **kwargs):

        obj = object.__new__(cls)

        # ensure self.comm is set, though usually already set by the child.
        obj.comm = kwargs.get('comm', CurrentMPIComm.get())

        # user-provided overrides and defaults for columns
        obj._overrides = {}

        # stores memory owner
        obj.base = None

        return obj

    def __finalize__(self, other):
        """
        Finalize the creation of a CatalogSource object by copying over
        any additional attributes from a second CatalogSource.

        The idea here is to only copy over attributes that are similar
        to meta-data, so we do not copy some of the core attributes of the
        :class:`CatalogSource` object.

        Parameters
        ----------
        other :
            the second object to copy over attributes from; it needs to be
            a subclass of CatalogSourcBase for attributes to be copied

        Returns
        -------
        CatalogSource :
            return ``self``, with the added attributes
        """
        if isinstance(other, CatalogSourceBase):
            d = other.__dict__.copy()
            nocopy = ['base', '_overrides', '_hardcolumns', '_defaults', 'comm',
                      '_size', '_csize']
            for key in d:
                if key not in nocopy:
                    self.__dict__[key] = d[key]

        return self

    def __iter__(self):
        return iter(self.columns)

    def __contains__(self, col):
        return col in self.columns

    def __slice__(self, index):
        """
        Select a subset of ``self`` according to a boolean index array.

        Returns a new object of the same type as ``selff`` holding only the
        data that satisfies the slice index.

        Parameters
        ----------
        index : array_like
            either a dask or numpy boolean array; this determines which
            rows are included in the returned object
        """
        # compute the index slice if needed and get the size
        if isinstance(index, da.Array):
            index = self.compute(index)
        elif isinstance(index, list):
            index = numpy.array(index)

        if getattr(self, 'size', NotImplemented) is NotImplemented:
            raise ValueError("cannot make catalog subset; self catalog doest not have a size")

        # verify the index is a boolean array
        if len(index) != self.size:
            raise ValueError("slice index has length %d; should be %d" %(len(index), self.size))
        if getattr(index, 'dtype', None) != '?':
            raise ValueError("index used to slice CatalogSource must be boolean and array-like")

        # new size is just number of True entries
        size = index.sum()

        # if collective size is unchanged, just return self
        if self.comm.allreduce(size) == self.csize:
           return self.base if self.base is not None else self

        # initialize subset Source of right size
        subset_data = {col:self[col][index] for col in self}
        cls = self.__class__ if self.base is None else self.base.__class__
        toret = cls._from_columns(size, self.comm, **subset_data)

        # attach the needed attributes
        toret.__finalize__(self)

        return toret

    def __getitem__(self, sel):
        """
        The following types of indexing are supported:

        #.  strings specifying a column in the CatalogSource; returns
            a dask array holding the column data
        #.  boolean arrays specifying a slice of the CatalogSource;
            returns a CatalogSource holding only the revelant slice
        #.  slice object specifying which particles to select
        #.  list of strings specifying column names; returns a CatalogSource
            holding only the selected columns

        Notes
        -----
        - Slicing with a boolean array is a **collective** operation
        - If the :attr:`base` attribute is set, columns will be returned
          from :attr:`base` instead of from ``self``.
        """
        # handle boolean array slices
        if not isinstance(sel, string_types):

            # size must be implemented
            if getattr(self, 'size', NotImplemented) is NotImplemented:
                raise ValueError("cannot perform selection due to NotImplemented size")

            # convert slices to boolean arrays
            if isinstance(sel, (slice, list)):

                # select a subset of list of string column names
                if isinstance(sel, list) and all(isinstance(ss, string_types) for ss in sel):
                    invalid = set(sel) - set(self.columns)
                    if len(invalid):
                        msg = "cannot select subset of columns from "
                        msg += "CatalogSource due to invalid columns: %s" % str(invalid)
                        raise KeyError(msg)

                    # return a CatalogSource only holding the selected columns
                    subset_data = {col:self[col] for col in sel}
                    toret = CatalogSource._from_columns(self.size, self.comm, **subset_data)
                    toret.attrs.update(self.attrs)
                    return toret

                # list must be all integers
                if isinstance(sel, list) and not numpy.array(sel).dtype == numpy.integer:
                    raise KeyError("array like indexing via a list should be a list of integers")

                # convert into slice into boolean array
                index = numpy.zeros(self.size, dtype='?')
                index[sel] = True; sel = index

            # do the slicing
            if not numpy.isscalar(sel):
                return self.__slice__(sel)
            else:
                raise KeyError("strings and boolean arrays are the only supported indexing methods")

        # owner of the memory (either self or base)
        memowner = self if self.base is None else self.base

        # get the right column
        is_default = False
        if sel in memowner._overrides:
            r = memowner._overrides[sel]
        elif sel in memowner.hardcolumns:
            r = memowner.get_hardcolumn(sel)
        elif sel in memowner._defaults:
            r = getattr(memowner, sel)()
            is_default = True
        else:
            raise KeyError("column `%s` is not defined in this source; " %sel + \
                            "try adding column via `source[column] = data`")

        # return a ColumnAccessor for pretty prints
        return ColumnAccessor(memowner, r, is_default=is_default)

    def __setitem__(self, col, value):
        """
        Add new columns to the CatalogSource, overriding any existing columns
        with the name ``col``.

        .. note::
            If the :attr:`base` attribute is set, columns will be added to
            :attr:`base` instead of to ``self``.
        """
        if self.base is not None: return self.base.__setitem__(col, value)

        self._overrides[col] = self.make_column(value)

    def __delitem__(self, col):
        """
        Delete a column; cannot delete a "hard-coded" column.

        .. note::
            If the :attr:`base` attribute is set, columns will be deleted
            from :attr:`base` instead of from ``self``.
        """
        if self.base is not None: return self.base.__delitem__(col)

        if col not in self.columns:
            raise ValueError("no such column, cannot delete it")

        if col in self.hardcolumns:
            raise ValueError("cannot delete a hard-coded column")

        if col in self._overrides:
            del self._overrides[col]
            return

        raise ValueError("unable to delete column '%s' from CatalogSource" %col)

    @property
    def attrs(self):
        """
        A dictionary storing relevant meta-data about the CatalogSource.
        """
        try:
            return self._attrs
        except AttributeError:
            self._attrs = {}
            return self._attrs

    @property
    def hardcolumns(self):
        """
        A list of the hard-coded columns in the CatalogSource.

        These columns are usually member functions marked by ``@column``
        decorator. Subclasses may override this method and use
        :func:`get_hardcolumn` to bypass the decorator logic.

        .. note::
            If the :attr:`base` attribute is set, the value of
            ``base.hardcolumns`` will be returned.
        """
        if self.base is not None: return self.base.hardcolumns

        # return the non-default, hard-coded columns, as determined by
        # ColumnFinder metaclass
        return sorted(self._hardcolumns)

    @property
    def columns(self):
        """
        All columns in the CatalogSource, including those hard-coded into
        the class's defintion and override columns provided by the user.

        .. note::
            If the :attr:`base` attribute is set, the value of
            ``base.columns`` will be returned.
        """
        if self.base is not None: return self.base.columns

        overrides = list(self._overrides)
        defaults = list(self._defaults)
        return sorted(set(self.hardcolumns + overrides + defaults))

    def copy(self):
        """
        Return a shallow copy of the object, where each column is a reference
        of the corresponding column in ``self``.

        .. note::
            No copy of data is made.

        .. note::
            This is different from view in that the attributes dictionary
            of the copy no longer related to ``self``.

        Returns
        -------
        CatalogSource :
            a new CatalogSource that holds all of the data columns of ``self``
        """
        # a new empty object with proper size
        toret = CatalogSourceBase.__new__(self.__class__, self.comm)
        toret._size = self.size
        toret._csize = self.csize

        # attach attributes from self and return
        toret = toret.__finalize__(self)

        # finally, add the data columns from self
        for col in self.columns:
            toret[col] = self[col]

        # copy the attributes too, so they become decoupled
        # this is different from view.
        toret._attrs = self._attrs.copy()

        return toret

    def get_hardcolumn(self, col):
        """
        Construct and return a hard-coded column.

        These are usually produced by calling member functions marked by the
        ``@column`` decorator.

        Subclasses may override this method and the hardcolumns attribute to
        bypass the decorator logic.

        .. note::
            If the :attr:`base` attribute is set, ``get_hardcolumn()``
            will called using :attr:`base` instead of ``self``.
        """
        if self.base is not None: return self.base.get_hardcolumn(col)

        if col in self._hardcolumns:
            return getattr(self, col)()
        else:
            raise ValueError("no such hard-coded column %s" %col)

    def compute(self, *args, **kwargs):
        """
        Our version of :func:`dask.compute` that computes
        multiple delayed dask collections at once.

        This should be called on the return value of :func:`read`
        to converts any dask arrays to numpy arrays.

        This uses the global cache as controlled by
        :class:`nbodykit.GlobalCache` to cache dask task computations.
        The default size is controlled by the ``global_cache_size`` global
        option; see :class:`set_options`. To set the size, see
        :func:`nbodykit.GlobalCache.resize`.

        .. note::
            If the :attr:`base` attribute is set, ``compute()``
            will called using :attr:`base` instead of ``self``.

        Parameters
        -----------
        args : object
            Any number of objects. If the object is a dask
            collection, it's computed and the result is returned.
            Otherwise it's passed through unchanged.

        Notes
        -----
        The dask default optimizer induces too many (unnecesarry)
        IO calls -- we turn this off feature off by default. Eventually we
        want our own optimizer probably.
        """
        import dask

        # return the base compute if it exists
        if self.base is not None:
            return self.base.compute(*args, **kwargs)

        # do not optimize graph (can lead to slower optimizations)
        kwargs.setdefault('optimize_graph', False)

        # compute using global cache
        with GlobalCache.get():
            toret = dask.compute(*args, **kwargs)

        # do not return tuples of length one
        if len(toret) == 1: toret = toret[0]
        return toret

    def save(self, output, columns, datasets=None, header='Header'):
        """
        Save the CatalogSource to a :class:`bigfile.BigFile`.

        Only the selected columns are saved and :attr:`attrs` are saved in
        ``header``. The attrs of columns are stored in the datasets.

        Parameters
        ----------
        output : str
            the name of the file to write to
        columns : list of str
            the names of the columns to save in the file
        datasets : list of str, optional
            names for the data set where each column is stored; defaults to
            the name of the column
        header : str, optional
            the name of the data set holding the header information, where
            :attr:`attrs` is stored
        """
        import bigfile
        import json
        from nbodykit.utils import JSONEncoder

        # trim out any default columns; these do not need to be saved as
        # they are automatically available to every Catalog
        columns = [col for col in columns if not self[col].is_default]

        # also make sure no default columns in datasets
        if datasets is None:
            datasets = columns
        else:
            datasets = [col for col in datasets if not self[col].is_default]

        if len(datasets) != len(columns):
            raise ValueError("`datasets` must have the same length as `columns`")

        with bigfile.BigFileMPI(comm=self.comm, filename=output, create=True) as ff:
            try:
                bb = ff.open(header)
            except:
                bb = ff.create(header)
            with bb :
                for key in self.attrs:
                    try:
                        bb.attrs[key] = self.attrs[key]
                    except ValueError:
                        try:
                            json_str = 'json://'+json.dumps(self.attrs[key], cls=JSONEncoder)
                            bb.attrs[key] = json_str
                        except:
                            raise ValueError("cannot save '%s' key in attrs dictionary" % key)

            for column, dataset in zip(columns, datasets):
                c = self[column]

                # save column attrs too
                with ff.create_from_array(dataset, c) as bb:
                    if hasattr(c, 'attrs'):
                        for key in c.attrs:
                            bb.attrs[key] = c.attrs[key]

    def read(self, columns):
        """
        Return the requested columns as dask arrays.

        Parameters
        ----------
        columns : list of str
            the names of the requested columns

        Returns
        -------
        list of :class:`dask.array.Array` :
            the list of column data, in the form of dask arrays
        """
        missing = set(columns) - set(self.columns)
        if len(missing) > 0:
            msg = "source does not contain columns: %s; " %str(missing)
            msg += "try adding columns via `source[column] = data`"
            raise ValueError(msg)

        return [self[col] for col in columns]

    def view(self, type=None):
        """
        Return a "view" of the CatalogSource object, with the returned
        type set by ``type``.

        This initializes a new empty class of type ``type`` and attaches
        attributes to it via the :func:`__finalize__` mechanism.

        Parameters
        ----------
        type : Python type
            the desired class type of the returned object.
        """
        # an empty class
        type = self.__class__ if type is None else type
        obj = CatalogSourceBase.__new__(type, self.comm)

        # propagate the size attributes
        obj._size = self.size
        obj._csize = self.csize

        # the new object's base points to self
        obj.base = self

        # attach the necessary attributes from self
        return obj.__finalize__(self)

    def to_mesh(self, Nmesh=None, BoxSize=None, dtype='f4', interlaced=False,
                compensated=False, window='cic', weight='Weight',
                value='Value', selection='Selection', position='Position'):
        """
        Convert the CatalogSource to a MeshSource, using the specified
        parameters.

        Parameters
        ----------
        Nmesh : int, optional
            the number of cells per side on the mesh; must be provided if
            not stored in :attr:`attrs`
        BoxSize : scalar, 3-vector, optional
            the size of the box; must be provided if
            not stored in :attr:`attrs`
        dtype : string, optional
            the data type of the mesh array
        interlaced : bool, optional
            use the interlacing technique of Sefusatti et al. 2015 to reduce
            the effects of aliasing on Fourier space quantities computed
            from the mesh
        compensated : bool, optional
            whether to correct for the window introduced by the grid
            interpolation scheme
        window : str, optional
            the string specifying which window interpolation scheme to use;
            see `pmesh.window.methods`
        weight : str, optional
            the name of the column specifying the weight for each particle
        value: str, optional
            the name of the column specifying the field value for each particle
        selection : str, optional
            the name of the column that specifies which (if any) slice
            of the CatalogSource to take
        position : str, optional
            the name of the column that specifies the position data of the
            objects in the catalog

        Returns
        -------
        mesh : CatalogMesh
            a mesh object that provides an interface for gridding particle
            data onto a specified mesh
        """
        from nbodykit.base.catalogmesh import CatalogMesh
        from pmesh.window import methods

        # make sure all of the columns exist
        for col in [weight, selection]:
            if col not in self:
                raise ValueError("column '%s' missing; cannot create mesh" %col)

        if window not in methods:
            raise ValueError("valid window methods: %s" %str(methods))

        if BoxSize is None:
            try:
                BoxSize = self.attrs['BoxSize']
            except KeyError:
                raise ValueError(("cannot convert particle source to a mesh; "
                                  "'BoxSize' keyword is not supplied and the CatalogSource "
                                  "does not define one in 'attrs'."))
        if Nmesh is None:
            try:
                Nmesh = self.attrs['Nmesh']
            except KeyError:
                raise ValueError(("cannot convert particle source to a mesh; "
                                  "'Nmesh' keyword is not supplied and the CatalogSource "
                                  "does not define one in 'attrs'."))

        return CatalogMesh(self, Nmesh=Nmesh,
                                 BoxSize=BoxSize,
                                 dtype=dtype,
                                 weight=weight,
                                 selection=selection,
                                 value=value,
                                 position=position,
                                 interlaced=interlaced,
                                 compensated=compensated,
                                 window=window)

class CatalogSource(CatalogSourceBase):
    """
    An abstract base class representing a catalog of discrete particles.

    This objects behaves like a structured numpy array -- it must have a
    well-defined size when initialized. The ``size`` here represents the
    number of particles in the source on the local rank.

    The information about each particle is stored as a series of
    columns in the format of dask arrays. These columns can be accessed
    in a dict-like fashion.

    All subclasses of this class contain the following default columns:

    #. ``Weight``
    #. ``Value``
    #. ``Selection``

    For a full description of these default columns, see
    :ref:`the documentation <catalog-source-default-columns>`.

    .. important::
        Subclasses of this class must set the ``_size`` attribute.

    Parameters
    ----------
    comm :
        the MPI communicator to use for this object
    """
    logger = logging.getLogger('CatalogSource')

    @classmethod
    def _from_columns(cls, size, comm, **columns):
        """
        An internal constructor to create a CatalogSource (or subclass)
        from a set of columns.

        This method is used internally by nbodykit to create
        views of catalogs based on existing catalogs.

        Use :class:`~nbodykit.source.catalog.array.ArrayCatalog` to adapt
        a structured array or dictionary to a CatalogSource.

        .. note::
            The returned object is of type ``cls`` and the only attributes
            set for the returned object are :attr:`size` and :attr:`csize`.
        """
        # the new empty object to return
        obj = CatalogSourceBase.__new__(cls, comm)

        # compute the sizes
        obj._size = size
        obj._csize = obj.comm.allreduce(obj._size)

        # add the columns in
        for name in columns:
            obj[name] = columns[name]

        return obj

    def __init__(self, *args, **kwargs):

        # if size is implemented, compute the csize
        if self.size is not NotImplemented:
            self._csize = self.comm.allreduce(self.size)
        else:
            self._csize = NotImplemented

    def __repr__(self):
        size = "%d" %self.size if self.size is not NotImplemented else "NotImplemented"
        return "%s(size=%s)" %(self.__class__.__name__, size)

    def __len__(self):
        """
        The local size of the CatalogSource on a given rank.
        """
        return self.size

    def __setitem__(self, col, value):
        """
        Add columns to the CatalogSource, overriding any existing columns
        with the name ``col``.
        """
        # handle scalar values
        if numpy.isscalar(value):
            assert self.size is not NotImplemented, "size is not implemented! cannot set scalar array"
            value = ConstantArray(value, self.size, chunks=_global_options['dask_chunk_size'])

        # check the correct size, if we know the size
        if self.size is not NotImplemented:
            args = (col, self.size, len(value))
            msg = "error setting '%s' column, data must be array of size %d, not %d" % args
            assert len(value) == self.size, msg

        # call the base __setitem__
        CatalogSourceBase.__setitem__(self, col, value)

    @property
    def size(self):
        """
        The number of objects in the CatalogSource on the local rank.

        If the :attr:`base` attribute is set, the ``base.size`` attribute
        will be returned.

        .. important::
            This property must be defined for all subclasses.
        """
        if self.base is not None: return self.base.size

        if not hasattr(self, '_size'):
            return NotImplemented
        return self._size

    @size.setter
    def size(self, value):
        raise RuntimeError(("Property size is read-only. Internally, ``_size`` "
                            "can be set during initialization."))

    @property
    def csize(self):
        """
        The total, collective size of the CatalogSource, i.e., summed across
        all ranks.

        It is the sum of :attr:`size` across all available ranks.

        If the :attr:`base` attribute is set, the ``base.csize`` attribute
        will be returned.
        """
        if self.base is not None: return self.base.csize

        if not hasattr(self, '_csize'):
            return NotImplemented
        return self._csize

    def gslice(self, start, stop, end=1, redistribute=True):
        """
        Execute a global slice of a CatalogSource.

        .. note::
            After the global slice is performed, the data is scattered
            evenly across all ranks.

        Parameters
        ----------
        start : int
            the start index of the global slice
        stop : int
            the stop index of the global slice
        step : int, optional
            the default step size of the global size
        redistribute : bool, optional
            if ``True``, evenly re-distribute the sliced data across all
            ranks, otherwise just return any local data part of the global
            slice
        """
        from nbodykit.utils import ScatterArray, GatherArray

        # determine the boolean index corresponding to the slice
        if self.comm.rank == 0:
            index = numpy.zeros(self.csize, dtype=bool)
            index[slice(start, stop, end)] = True
        else:
            index = None
        index = self.comm.bcast(index)

        # scatter the index back to all ranks
        counts = self.comm.allgather(self.size)
        index = ScatterArray(index, self.comm, root=0, counts=counts)

        # perform the needed local slice
        subset = self[index]

        # if we don't want to redistribute evenly, just return the slice
        if not redistribute:
            return subset

        # re-distribute each column from the sliced data
        # NOTE: currently Gather/ScatterArray requires numpy arrays, but
        # in principle we could pass dask arrays around between ranks and
        # avoid compute() calls
        data = self.compute(*[subset[col] for col in subset])

        # gather/scatter each column
        evendata = {}
        for i, col in enumerate(subset):
            alldata = GatherArray(data[i], self.comm, root=0)
            evendata[col] = ScatterArray(alldata, self.comm, root=0)

        # return a new CatalogSource holding the evenly distributed data
        size = len(evendata[col])
        toret = self.__class__._from_columns(size, self.comm, **evendata)
        return toret.__finalize__(self)

    def sort(self, keys, reverse=False, usecols=None):
        """
        Return a CatalogSource, sorted globally across all MPI ranks
        in ascending order by the input keys.

        Sort columns must be floating or integer type.

        .. note::
            After the sort operation, the data is scattered evenly across
            all ranks.

        Parameters
        ----------
        keys : list, tuple
            the names of columns to sort by. If multiple columns are provided,
            the data is sorted consecutively in the order provided
        reverse : bool, optional
            if ``True``, perform descending sort operations
        usecols : list, optional
            the name of the columns to include in the returned CatalogSource
        """
        # single string passed as input
        if isinstance(keys, string_types):
            keys = [keys]

        # no duplicated keys
        if len(set(keys)) != len(keys):
            raise ValueError("duplicated sort keys")

        # all keys must be valid
        bad = set(keys) - set(self.columns)
        if len(bad):
            raise ValueError("invalid sort keys: %s" %str(bad))

        # check usecols input
        if usecols is not None:
            if isinstance(usecols, string_types):
                usecols = [usecols]
            if not isinstance(usecols, (list, tuple)):
                raise ValueError("usecols should be a list or tuple of column names")

            bad = set(usecols) - set(self.columns)
            if len(bad):
                raise ValueError("invalid column names in usecols: %s" %str(bad))

        # sort the data
        data = _sort_data(self.comm, self, list(keys), reverse=reverse, usecols=usecols)

        # get a dictionary of data
        cols = {}
        for col in data.dtype.names:
            cols[col] = data[col]

        # make the new CatalogSource from the data dict
        size = len(data)

        # NOTE: if we are slicing by columns too, we must return a bare CatalogSource
        # we cannot guarantee the consistency of the hard column definitions otherwise
        if usecols is None:
            toret = self.__class__._from_columns(size, self.comm, **cols)
            return toret.__finalize__(self)
        else:
            toret = CatalogSource._from_columns(size, self.comm, **cols)
            toret.attrs.update(self.attrs)
            return toret

    @column(is_default=True)
    def Selection(self):
        """
        A boolean column that selects a subset slice of the CatalogSource.

        By default, this column is set to ``True`` for all particles, and
        all CatalogSource objects will contain this column.
        """
        return ConstantArray(True, self.size, chunks=_global_options['dask_chunk_size'])

    @column(is_default=True)
    def Weight(self):
        """
        The column giving the weight to use for each particle on the mesh.

        The mesh field is a weighted average of ``Value``, with the weights
        given by ``Weight``.

        By default, this array is set to unity for all particles, and
        all CatalogSource objects will contain this column.
        """
        return ConstantArray(1.0, self.size, chunks=_global_options['dask_chunk_size'])

    @property
    def Index(self):
        """
        The attribute giving the global index rank of each particle in the
        list. It is an integer from 0 to ``self.csize``.

        Note that slicing changes this index value.
        """
        offset = sum(self.comm.allgather(self.size)[:self.comm.rank])
        # do not use u8, because many numpy casting rules case u8 to f8 automatically.
        # it is ridiculous.
        return da.arange(offset, offset + self.size, dtype='i8',
               chunks=_global_options['dask_chunk_size'])

    @column(is_default=True)
    def Value(self):
        """
        When interpolating a CatalogSource on to a mesh, the value of this
        array is used as the Value that each particle contributes to a given
        mesh cell.

        The mesh field is a weighted average of ``Value``, with the weights
        given by ``Weight``.

        By default, this array is set to unity for all particles, and
        all CatalogSource objects will contain this column.
        """
        return ConstantArray(1.0, self.size, chunks=_global_options['dask_chunk_size'])


def _sort_data(comm, cat, rankby, reverse=False, usecols=None):
    """
    Sort the input data by the specified columns

    Parameters
    ----------
    comm :
        the mpi communicator
    cat : CatalogSource
        the catalog holding the data to sort
    rankby : list of str
        list of columns to sort by
    reverse : bool, optional
        if ``True``, sort in descending order
    usecols : list, optional
        only sort these data columns
    """
    import mpsort

    # determine which columns we need
    if usecols is None:
        usecols = cat.columns

    # remove duplicates from usecols
    usecols = list(set(usecols))

    # the columns we need in the sort steps
    columns = list(set(rankby)|set(usecols))

    # make the data to sort
    dtype = [('_sortkey', 'i8')]
    for col in cat:
        if col in columns:
            dt = (cat[col].dtype.char,)
            dt += cat[col].shape[1:]
            if len(dt) == 1: dt = dt[0]
            dtype.append((col, dt))
    dtype = numpy.dtype(dtype)

    data = numpy.empty(cat.size, dtype=dtype)
    for col in columns:
        data[col] = cat[col]

    # sort the particles by the specified columns and store the
    # corrected sorted index
    for col in reversed(rankby):
        dt = data.dtype[col]
        rankby_name = col

        # make an integer key for floating columns
        if issubclass(dt.type, numpy.floating):
            data['_sortkey'] = numpy.fromstring(data[col].tobytes(), dtype='i8')
            if reverse:
                data['_sortkey'] *= -1
            rankby_name = '_sortkey'
        elif not issubclass(dt.type, numpy.integer):
            args = (col, str(dt))
            raise ValueError("cannot sort by column '%s' with dtype '%s'; must be integer or floating type" %args)

        # do the parallel sort
        mpsort.sort(data, orderby=rankby_name, comm=comm)

    return data[usecols]



from nbodykit.base.catalog import CatalogSource, column
from nbodykit import CurrentMPIComm

from numpy.random import RandomState
import numpy
import functools
import contextlib

N_PER_SEED = 100000

def _mpi_enabled_rng(func):
    """
    A decorator that handles generating random numbers in parallel
    in a manner that is independent of the number of ranks

    Designed to be used with :class:`MPIRandomState`
    """
    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):

        self = func.__self__
        size = kwargs.get('size', None)
        if size is not None and isinstance(size, int):
            size = (size,)

        # do nothing if size not provided or wrong
        if size is None or size[0] != self.size:
            return func(*args, **kwargs)

        # size matches the "size" attribute
        else:
            toret = []

            # loop through chunks that this rank is responsible for
            for chunk, (start, stop) in zip(self._chunks, self._slices):
                with self.seeded_context(self._seeds[chunk]):

                    kwargs['size'] = (N_PER_SEED,) + size[1:]
                    if chunk == self.N // N_PER_SEED:
                        kwargs['size'] = (self.N % N_PER_SEED,) + size[1:]
                    toret.append(func(*args, **kwargs)[start:stop])
            return numpy.concatenate(toret, axis=0)

    func_wrapper.mpi_enabled = True
    return func_wrapper

class MPIRandomState(RandomState):
    """
    A wrapper around :class:`numpy.random.RandomState` that can return
    random numbers in parallel, independent of the number of ranks.

    Parameters
    ----------
    comm : MPI communicator
        the MPI communicator
    seed : int
        the global seed that seeds all other random seeds
    N : int
        the total size of the random numbers to generate; we return chunks of
        the total on each CPU, based on the CPU's rank
    """
    def __init__(self, comm, seed, N):

        RandomState.__init__(self, seed=seed)

        # the number of seeds to generate N particles
        n_seeds = N // N_PER_SEED
        if N % N_PER_SEED: n_seeds += 1

        self.comm        = comm
        self.global_seed = seed
        self.N           = N

        start = N * comm.rank // comm.size
        stop  = N * (comm.rank  + 1) // comm.size
        self.size  = stop - start

        # generate the full set of seeds from the global seed
        rng = numpy.random.RandomState(seed=seed)
        self._seeds = rng.randint(0, high=0xffffffff, size=n_seeds)

        # sizes of each chunk
        sizes = [N_PER_SEED]*(N//N_PER_SEED)
        if N % N_PER_SEED: sizes.append(N % N_PER_SEED)

        # the local chunks this rank is responsible for
        cumsizes = numpy.insert(numpy.cumsum(sizes), 0, 0)
        chunk_range = numpy.searchsorted(cumsizes[1:], [start, stop])
        self._chunks = list(range(chunk_range[0], chunk_range[1]+1))

        # and the slices for each local chunk
        self._slices = []
        for chunk in self._chunks:
            start_size = cumsizes[chunk]
            sl = (max(start-start_size, 0), min(stop-start_size, sizes[chunk]))
            self._slices.append(sl)

    def __dir__(self):
        """
        Explicitly set the attributes as those of the RandomState class too
        """
        d1 = set(RandomState().__dir__())
        d2 = set(RandomState.__dir__(self))
        return list(d1|d2)

    def __getattribute__(self, name):
        """
        Decorate callable functions of RandomState such that they return
        chunks of the total `N` random numbers generated
        """
        attr = RandomState.__getattribute__(self, name)
        if callable(attr) and not getattr(attr, 'mpi_enabled', False):
            attr = _mpi_enabled_rng(attr)
        return attr

    @contextlib.contextmanager
    def seeded_context(self, seed):
        """
        A context manager to set and then restore the random seed
        """
        startstate = self.get_state()
        self.seed(seed)
        yield
        self.set_state(startstate)

class RandomCatalog(CatalogSource):
    """
    A CatalogSource that can have columns added via a
    collective random number generator.

    The random number generator stored as :attr:`rng` behaves
    as :class:`numpy.random.RandomState` but generates random
    numbers only on the local rank in a manner independent of
    the number of ranks.

    Parameters
    ----------
    csize : int
        the desired collective size of the Source
    seed : int, optional
        the global seed for the random number generator
    comm : MPI communicator
        the MPI communicator; set automatically if None
    """
    def __repr__(self):
        args = (self.size, self.attrs['seed'])
        return "RandomCatalog(size=%d, seed=%s)" % args

    @CurrentMPIComm.enable
    def __init__(self, csize, seed=None, comm=None):

        self.comm = comm

        # set the seed randomly if it is None
        if seed is None:
            if self.comm.rank == 0:
                seed = numpy.random.randint(0, 4294967295)
            seed = self.comm.bcast(seed)
        self.attrs['seed'] = seed

        # generate the seeds from the global seed
        if csize == 0:
            raise ValueError("no random particles generated!")
        self._rng = MPIRandomState(comm, seed, csize)
        self._size =  self.rng.size

        # init the base class
        CatalogSource.__init__(self, comm=comm)

    @property
    def rng(self):
        """
        A :class:`MPIRandomState` that behaves as
        :class:`numpy.random.RandomState` but generates random
        numbers in a manner independent of the number of ranks.
        """
        return self._rng

class UniformCatalog(RandomCatalog):
    """
    A CatalogSource that has uniformly-distributed ``Position``
    and ``Velocity`` columns.

    The random numbers generated do not depend on the number of
    available ranks.

    Parameters
    ----------
    nbar : float
        the desired number density of particles in the box
    BoxSize : float, 3-vector
        the size of the box
    seed : int, optional
        the random seed
    comm :
        the MPI communicator
    """
    def __repr__(self):
        args = (self.size, self.attrs['seed'])
        return "UniformCatalog(size=%d, seed=%s)" % args

    @CurrentMPIComm.enable
    def __init__(self, nbar, BoxSize, seed=None, comm=None):

        self.comm    = comm

        _BoxSize = numpy.empty(3, dtype='f8')
        _BoxSize[:] = BoxSize
        self.attrs['BoxSize'] = _BoxSize

        rng = numpy.random.RandomState(seed)
        N = rng.poisson(nbar * numpy.prod(self.attrs['BoxSize']))
        if N == 0:
            raise ValueError("no uniform particles generated, try increasing `nbar` parameter")
        RandomCatalog.__init__(self, N, seed=seed, comm=comm)

        self._pos = self.rng.uniform(size=(self._size, 3)) * self.attrs['BoxSize']
        self._vel = self.rng.uniform(size=(self._size, 3)) * self.attrs['BoxSize'] * 0.01

    @column
    def Position(self):
        """
        The position of particles, uniformly distributed in :attr:`BoxSize`
        """
        return self.make_column(self._pos)

    @column
    def Velocity(self):
        """
        The velocity of particles, uniformly distributed in ``0.01 x BoxSize``
        """
        return self.make_column(self._vel)



# FFT-based
from .fftpower import FFTPower, ProjectedFFTPower
from .fftcorr import FFTCorr
from .convpower import ConvolvedFFTPower

# grouping
from .fof import FOF
from .fibercollisions import FiberCollisions
from .cgm import CylindricalGroups

# pair counters, correlation functions
from .pair_counters import SurveyDataPairCount, SimulationBoxPairCount
from .paircount_tpcf import SurveyData2PCF, SimulationBox2PCF
from .threeptcf import SimulationBox3PCF, SurveyData3PCF

# miscellaneous
from .kdtree import KDDensity
from .zhist import RedshiftHistogram

__all__ = ['FFTPower',
           'ProjectedFFTPower',
           'FFTCorr',
           'ConvolvedFFTPower',
           'FOF',
           'FiberCollisions',
           'CylindricalGroups',
           'SurveyDataPairCount',
           'SurveyData2PCF',
           'SurveyData3PCF',
           'SimulationBoxPairCount',
           'SimulationBox2PCF',
           'SimulationBox3PCF',
           'KDDensity',
           'RedshiftHistogram',
          ]



from .bigfile import BigFileMesh
from .linear import LinearMesh
from .field import FieldMesh
from .array import ArrayMesh

__all__ = ['BigFileMesh',
           'LinearMesh',
           'FieldMesh',
           'ArrayMesh',
          ]



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_array_equal, assert_allclose
import pytest

setup_logging()

@MPITest([1, 4])
def test_paint(comm):

    CurrentMPIComm.set(comm)

    NBAR1 = 3e-5; WEIGHT1 = 1.05
    NBAR2 = 3e-3; WEIGHT2 = 0.95
    P0_FKP = 2e4

    # the catalog
    source1 = UniformCatalog(nbar=NBAR1, BoxSize=512., seed=42)
    source2 = UniformCatalog(nbar=NBAR2, BoxSize=512., seed=84)

    # add completeness weights
    source1['Weight'] = WEIGHT1
    source2['Weight'] = WEIGHT2

    # add n(z)
    source1['NZ'] = NBAR1
    source2['NZ'] = NBAR2

    # add FKP weights
    FKP_WEIGHT1 = 1.0 / (1 + NBAR1*P0_FKP)
    FKP_WEIGHT2 = 1.0 / (1 + NBAR2*P0_FKP)
    source1['FKPWeight'] = FKP_WEIGHT1
    source2['FKPWeight'] = FKP_WEIGHT2

    # initialize the catalog
    cat = FKPCatalog(source1, source2)

    # the meshes
    mesh = cat.to_mesh(Nmesh=32, BoxSize=512)
    mesh1 = source1.to_mesh(Nmesh=32, BoxSize=512)
    mesh2 = source2.to_mesh(Nmesh=32, BoxSize=512)

    # update weights for source1 and source2
    mesh1['Weight'] *= mesh1['FKPWeight']
    mesh2['Weight'] *= mesh2['FKPWeight']

    # paint the re-centered Position
    mesh1['Position'] -= mesh.attrs['BoxCenter']
    mesh2['Position'] -= mesh.attrs['BoxCenter']

    # alpha is the sum of Weight
    alpha = 1. * source1.csize * WEIGHT1 / (source2.csize * WEIGHT2)

    # paint
    real1 = mesh1.to_real_field(normalize=False)
    real2 = mesh2.to_real_field(normalize=False)

    # the FKP density field: n_data - alpha*n_randoms
    vol_per_cell = (512.0/32)**3
    fkp_density = (real1.value - alpha*real2.value)/vol_per_cell

    # from the FKPCatalog
    combined = mesh.to_real_field()

    # check meta-data
    assert_allclose(combined.attrs['alpha'], alpha)

    # total number
    assert_allclose(combined.attrs['data.N'], source1.csize)
    assert_allclose(combined.attrs['randoms.N'], source2.csize)

    # total weighted number
    assert_allclose(combined.attrs['data.W'], source1.csize*WEIGHT1)
    assert_allclose(combined.attrs['randoms.W'], source2.csize*WEIGHT2)

    # must be the same
    assert_allclose(combined.value, fkp_density, atol=1e-5)



from runtests.mpi import MPITest
from nbodykit.io.csv import CSVFile

import os
import numpy
import tempfile
import pytest

@MPITest([1])
def test_dask(comm):

    with tempfile.NamedTemporaryFile() as ff:    
        
        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)
        
        # read
        names =['a', 'b', 'c', 'd', 'e']
        f = CSVFile(path=ff.name, names=names, blocksize=100)
        
        a_dask = f.get_dask('a')
        numpy.testing.assert_almost_equal(a_dask.compute(), data[:,0])

@MPITest([1])
def test_getitem(comm):

    with tempfile.NamedTemporaryFile() as ff:    
        
        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)
        
        # read
        names =['a', 'b', 'c', 'd', 'e']
        f = CSVFile(path=ff.name, names=names, blocksize=100)

        # empty slice        
        with pytest.raises(IndexError):
            empty = f[[]]

        # cannot slice twice
        a = f['a']
        with pytest.raises(IndexError):
            a2 = a['a']
        
        # bad column name
        with pytest.raises(IndexError):
            bad = f[['BAD1', 'BAD2']]
        
        # slice multiple columns
        f2 = f[['a', 'b']]
        assert f2.columns == ['a', 'b']
        f3 = f2[['a']]
        assert f3.columns == ['a']
        
        # slice as an array 
        d = f.asarray()
        numpy.testing.assert_almost_equal(d[:,0:2], data[:,0:2])
        numpy.testing.assert_almost_equal(d[:50], data[:50])
        numpy.testing.assert_almost_equal(d[0].squeeze(), data[0])
        
        # pass list of integers
        numpy.testing.assert_almost_equal(d[[0,1,2]], data[[0,1,2]])
        
        # pass boolean slice
        valid = numpy.random.choice([True, False], replace=True, size=len(f))
        f2 = f[valid]
        numpy.testing.assert_almost_equal(f2['a'][:], data[valid, 0])
        
        # wrong slice shape
        with pytest.raises(IndexError):
            d2 = d[:,:,:]

        
@MPITest([1])
def test_asarray(comm):

    with tempfile.NamedTemporaryFile() as ff:    
        
        # generate data
        data = numpy.random.random(size=(100,5))
        numpy.savetxt(ff, data, fmt='%.7e'); ff.seek(0)
        
        # read
        names =['a', 'b', 'c', 'd', 'e']
        f = CSVFile(path=ff.name, names=names, dtype={'a':'f4', 'b':'f8'}, blocksize=100)
        
        # cannot do asarray twice
        a = f['a']
        with pytest.raises(ValueError):
            a2 = a.asarray()
        
        # cannot do view with different dtypes
        with pytest.raises(ValueError):
            f2 = f.asarray()



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from scipy.interpolate import InterpolatedUnivariateSpline
from numpy.testing import assert_allclose, assert_array_equal
import pytest

setup_logging("debug")

NDATA = 1000
NBAR = 1e-4

def make_sources(cosmo):

    data = RandomCatalog(NDATA, seed=42)
    randoms = RandomCatalog(NDATA*10, seed=84)

    # add the random columns
    for s in [data, randoms]:

        # ra, dec, z
        s['z']   = s.rng.normal(loc=0.5, scale=0.1, size=s.size)
        s['ra']  = s.rng.uniform(low=110, high=260, size=s.size)
        s['dec'] = s.rng.uniform(low=-3.6, high=60., size=s.size)

        # position
        s['Position'] = transform.SkyToCartesian(s['ra'], s['dec'], s['z'], cosmo=cosmo)

    return data, randoms

@MPITest([1, 4])
def test_diff_cross_boxsizes(comm):

    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo)
    for s in [data, randoms]:

        # constant number density
        s['NZ'] = NBAR

        # completeness weights
        P0 = 1e4
        s['Weight'] = (1 + P0*s['NZ'])**2

    # two mesh objects from same FKP source
    fkp = FKPCatalog(data, randoms)
    mesh1 = fkp.to_mesh(Nmesh=128, dtype='f8', nbar='NZ')

    # second mesh has larger box size
    mesh2 = fkp.to_mesh(Nmesh=128, dtype='f8', nbar='NZ', BoxSize=1.1*mesh1.attrs['BoxSize'])

    # compute the multipoles
    r = ConvolvedFFTPower(mesh1, second=mesh2, poles=[0,2,4], dk=0.005, use_fkp_weights=True, P0_FKP=P0)

    # make sure we matched the box sizes
    assert_array_equal(r.first.attrs['BoxSize'], r.second.attrs['BoxSize'])
    assert_array_equal(r.first.attrs['BoxCenter'], r.second.attrs['BoxCenter'])

@MPITest([1, 4])
def test_true_cross_corr_fail(comm):

    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo)
    for s in [data, randoms]:
        s['NZ'] = NBAR


    # two mesh objects from different FKP source
    fkp1 = FKPCatalog(data, randoms)
    fkp2 = FKPCatalog(data, randoms)

    mesh1 = fkp1.to_mesh(Nmesh=128, dtype='f8', nbar='NZ')
    mesh2 = fkp2.to_mesh(Nmesh=128, dtype='f8', nbar='NZ')

    # cannot do cross correlations with different data/randoms catalogs
    with pytest.raises(NotImplementedError):
        r = ConvolvedFFTPower(mesh1, second=mesh2, poles=[0,2,4], dk=0.005)


@MPITest([1, 4])
def test_bad_cross_corr_columns(comm):

    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo)
    for s in [data, randoms]:
        s['NZ'] = NBAR

    # same FKP source but different selection columns won't work!
    fkp = FKPCatalog(data, randoms)
    fkp['data']['Selection2'] = fkp['data/Selection']
    fkp['randoms']['Selection2'] = fkp['randoms/Selection']

    mesh1 = fkp.to_mesh(Nmesh=128, dtype='f8', nbar='NZ', selection='Selection')
    mesh2 = fkp.to_mesh(Nmesh=128, dtype='f8', nbar='NZ', selection='Selection2')

    # columns in mesh must be same except for weight
    with pytest.raises(NotImplementedError):
        r = ConvolvedFFTPower(mesh1, second=mesh2, poles=[0,2,4], dk=0.005)

@MPITest([1, 4])
def test_cross_corr(comm):

    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo)
    for s in [data, randoms]:

        # constant number density
        s['NZ'] = NBAR

        # completeness weights
        P0 = 1e4
        s['Weight'] = (1 + P0*s['NZ'])**2

    # two mesh objects from same FKP source
    fkp = FKPCatalog(data, randoms)
    mesh1 = fkp.to_mesh(Nmesh=128, dtype='f8', nbar='NZ', fkp_weight='FKPWeight', comp_weight='Weight', selection='Selection')
    mesh2 = fkp.to_mesh(Nmesh=128, dtype='f8', nbar='NZ', fkp_weight='FKPWeight', comp_weight='Weight', selection='Selection')

    # compute the multipoles
    r = ConvolvedFFTPower(mesh1, second=mesh2, poles=[0,2,4], dk=0.005, use_fkp_weights=True, P0_FKP=P0)

    # normalization
    assert_allclose(r.attrs['data.norm'], NDATA*NBAR)
    assert_allclose(r.attrs['randoms.norm'], NDATA*NBAR)

    # shotnoise
    S_data = r.attrs['first.data.W']/r.attrs['randoms.norm']
    S_ran = r.attrs['first.randoms.W']/r.attrs['randoms.norm']*r.attrs['alpha']**2
    S = S_data + S_ran
    assert_allclose(S, r.attrs['shotnoise'])

@MPITest([1, 4])
def test_bad_input(comm):

    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15

    # source has wrong type
    data, randoms = make_sources(cosmo)
    for s in [data, randoms]:
        s['NZ'] = NBAR
        s['FKPWeight'] = 1.0 / (1 + 2e4*s['NZ'])

    with pytest.raises(TypeError):
        r = ConvolvedFFTPower(data, poles=[0,2,4], dk=0.005, Nmesh=64)

    # the FKP source
    fkp = FKPCatalog(data, randoms)

    # must specify P0_FKP
    with pytest.raises(ValueError):
        r = ConvolvedFFTPower(fkp, poles=0, dk=0.005, use_fkp_weights=True, P0_FKP=None, Nmesh=64)

    # warn about overwriting FKP Weights
    with pytest.warns(UserWarning):
        r = ConvolvedFFTPower(fkp, poles=0, dk=0.005, use_fkp_weights=True, P0_FKP=1e4, Nmesh=64)


@MPITest([4])
def test_no_monopole(comm):

    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo)

    # select in given redshift range
    for s in [data, randoms]:
        s['NZ'] = NBAR

    # the FKP source
    fkp = FKPCatalog(data, randoms)
    fkp = fkp.to_mesh(Nmesh=128, dtype='f8', nbar='NZ')

    # compute the multipoles
    r = ConvolvedFFTPower(fkp, poles=[2,4], dk=0.005)

    assert 'power_0' not in r.poles.variables
    assert 'power_2' in r.poles.variables
    assert 'power_4' in r.poles.variables

@MPITest([4])
def test_bad_normalization(comm):

    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo)

    # select in given redshift range
    for s in [data, randoms]:
        s['NZ'] = NBAR

    # bad normalization via NZ
    randoms['NZ'] *= 50.0

    # the FKP source
    fkp = FKPCatalog(data, randoms)
    fkp = fkp.to_mesh(Nmesh=128, dtype='f8', nbar='NZ')

    # compute the multipoles
    with pytest.raises(ValueError):
        r = ConvolvedFFTPower(fkp, poles=[0,2,4], dk=0.005)

@MPITest([4])
def test_selection(comm):

    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo)

    # select in given redshift range
    for s in [data, randoms]:
        s['NZ'] = NBAR
        s['Selection'] = (s['z'] > 0.4)&(s['z'] < 0.6)

    # the FKP source
    fkp = FKPCatalog(data, randoms)
    fkp = fkp.to_mesh(Nmesh=128, dtype='f8', nbar='NZ', selection='Selection')

    # compute the multipoles
    r = ConvolvedFFTPower(fkp, poles=[0,2,4], dk=0.005)

    # number of data objects selected
    N = comm.allreduce(((data['z'] > 0.4)&(data['z'] < 0.6)).sum())
    assert_allclose(r.attrs['data.N'], N)

    # number of randoms selected
    N = comm.allreduce(((randoms['z'] > 0.4)&(randoms['z'] < 0.6)).sum())
    assert_allclose(r.attrs['randoms.N'], N)

    # and save
    r.save("conv-power-with-selection.json")

    # load and check output
    r2 = ConvolvedFFTPower.load("conv-power-with-selection.json")
    assert_array_equal(r.poles.data, r2.poles.data)

@MPITest([1, 4])
def test_run(comm):

    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo)
    for s in [data, randoms]:

        # constant number density
        s['NZ'] = NBAR

        # completeness weights
        P0 = 1e4
        s['Weight'] = (1 + P0*s['NZ'])**2

    # the FKP source
    fkp = FKPCatalog(data, randoms)
    fkp = fkp.to_mesh(Nmesh=128, dtype='f8', nbar='NZ', fkp_weight='FKPWeight', comp_weight='Weight', selection='Selection')

    # compute the multipoles
    r = ConvolvedFFTPower(fkp, poles=[0,2,4], dk=0.005, use_fkp_weights=True, P0_FKP=P0)

    # compute pkmu
    mu_edges = numpy.linspace(0, 1, 6)
    pkmu = r.to_pkmu(mu_edges=mu_edges, max_ell=4)

    # normalization
    assert_allclose(r.attrs['data.norm'], NDATA*NBAR)
    assert_allclose(r.attrs['randoms.norm'], NDATA*NBAR)

    # shotnoise
    S_data = r.attrs['data.W']/r.attrs['randoms.norm']
    S_ran = r.attrs['randoms.W']/r.attrs['randoms.norm']*r.attrs['alpha']**2
    S = S_data + S_ran
    assert_allclose(S, r.attrs['shotnoise'])


@MPITest([1, 4])
def test_with_zhist(comm):

    NDATA = 1000
    NBAR = 1e-4
    FSKY = 0.15

    CurrentMPIComm.set(comm)
    cosmo = cosmology.Planck15

    # make the sources
    data, randoms = make_sources(cosmo)

    # initialize the FKP source
    fkp = FKPCatalog(data, randoms)

    # compute NZ from randoms
    zhist = RedshiftHistogram(randoms, FSKY, cosmo, redshift='z')

    # add n(z) from randoms to the FKP source
    nofz = InterpolatedUnivariateSpline(zhist.bin_centers, zhist.nbar)
    fkp['randoms']['NZ'] = nofz(randoms['z'])
    fkp['data']['NZ'] = nofz(data['z'])

    # normalize NZ to the total size of the data catalog
    alpha = 1.0 * data.csize / randoms.csize
    fkp['randoms']['NZ'] *= alpha
    fkp['data']['NZ'] *= alpha

    # compute the multipoles
    r = ConvolvedFFTPower(fkp.to_mesh(Nmesh=128), poles=[0,2,4], dk=0.005)

    assert_allclose(r.attrs['data.norm'], 0.000388338522187, rtol=1e-5)
    assert_allclose(r.attrs['randoms.norm'], 0.000395808747269, rtol=1e-5)



import numpy
import mcfit
from scipy.interpolate import InterpolatedUnivariateSpline
from .power.zeldovich import ZeldovichPower

NUM_PTS = 1024

def xi_to_pk(r, xi, extrap=False):
    r"""
    Return a callable function returning the power spectrum, as computed
    from the Fourier transform of the input :math:`r` and :math:`\xi(r)` arrays.

    This uses the :mod:`mcfit` package perform the FFT.

    Parameters
    ----------
    r : array_like
        separation values where ``xi`` is evaluated
    xi : array_like
        the array holding the correlation function values
    extrap : bool, optional
        whether to extrapolate the power spectrum with a power law; can improve
        the smoothness of the FFT

    Returns
    -------
    InterpolatedUnivariateSpline :
        a spline holding the interpolated power spectrum values
    """
    P = mcfit.xi2P(r)
    kk, Pk = P(xi, extrap=extrap)
    return InterpolatedUnivariateSpline(kk, Pk)


def pk_to_xi(k, Pk, extrap=True):
    r"""
    Return a callable function returning the correlation function, as computed
    from the Fourier transform of the input :math:`k` and :math:`P(k)` arrays.

    This uses the :mod:`mcfit` package perform the FFT.

    Parameters
    ----------
    k : array_like
        wavenumbers where ``Pk`` is evaluated
    Pk : array_like
        the array holding the power spectrum values
    extrap : bool, optional
        whether to extrapolate the power spectrum with a power law; can improve
        the smoothness of the FFT

    Returns
    -------
    InterpolatedUnivariateSpline :
        a spline holding the interpolated correlation function values
    """
    xi = mcfit.P2xi(k)
    rr, CF = xi(Pk, extrap=extrap)
    return InterpolatedUnivariateSpline(rr, CF)


class CorrelationFunction(object):
    """
    Evaluate the correlation function by Fourier transforming
    a power spectrum object, with automatic re-scaling with redshift and sigma8.

    Parameters
    ----------
    power : callable
         a callable power spectrum that returns the power at a given ``k``;
         this should have ``redshift``, ``sigma8``, and ``cosmo`` attributes
    """
    def __init__(self, power):

        self.power = power

        # check required attributes
        for attr in ['redshift', 'sigma8', 'cosmo']:
            if not hasattr(power, attr):
                raise AttributeError("input power spectrum object must have a ``%s`` attribute" %attr)

        # meta-data
        self._attrs = {}
        self._attrs.update(getattr(self.power, 'attrs', {}))

    @property
    def attrs(self):
        self._attrs['redshift'] = self.redshift
        self._attrs['sigma8'] = self.sigma8
        return self._attrs

    @property
    def redshift(self):
        return self.power.redshift

    @redshift.setter
    def redshift(self, value):
        self.power.redshift = value

    @property
    def sigma8(self):
        return self.power.sigma8

    @sigma8.setter
    def sigma8(self, value):
        self.power.sigma8 = value

    @property
    def cosmo(self):
        return self.power.cosmo

    def __call__(self, r, smoothing=0., kmin=1e-5, kmax=10.):
        """
        Return the correlation function (dimensionless) for separations ``r``

        Parameters
        ----------
        r : float, array_like
            the separation array in units of :math:`h^{-1} \mathrm(Mpc)`
        smoothing  : float, optional
            the std deviation of the Fourier space Gaussian smoothing to apply
            to P(k) before taking the FFT
        kmin : float, optional
            the minimum ``k`` value to compute P(k) for before taking the FFT
        kmax : float, optional
            the maximum ``k`` value to compute P(k) for before taking the FFT
        """
        k = numpy.logspace(numpy.log10(kmin), numpy.log10(kmax), NUM_PTS)

        # power with smoothing
        Pk = self.power(k)
        Pk *= numpy.exp(-(k*smoothing)**2)

        # only extrap if not zeldovich
        extrap = not isinstance(self.power, ZeldovichPower)
        return pk_to_xi(k, Pk, extrap=extrap)(r)



import numpy
import numbers
from mpi4py import MPI

from pmesh.pm import RealField, ComplexField
from nbodykit.meshtools import SlabIterator
from nbodykit.utils import GatherArray, ScatterArray

def gaussian_complex_fields(pm, linear_power, seed,
            unitary_amplitude=False, inverted_phase=False,
            compute_displacement=False):
    r"""
    Make a Gaussian realization of a overdensity field, :math:`\delta(x)`.

    If specified, also compute the corresponding 1st order Lagrangian
    displacement field (Zel'dovich approximation) :math:`\psi(x)`,
    which is related to the linear velocity field via:

    .. math::

        v(x) = f a H \psi(x)

    Notes
    -----
    This computes the overdensity field using the following steps:

    #. Generate complex variates with unity variance
    #. Scale the Fourier field by :math:`(P(k) / V)^{1/2}`

    After step 2, the complex field has unity variance. This
    is equivalent to generating real-space normal variates
    with mean and unity variance, calling r2c() and dividing by :math:`N^3`
    since the variance of the complex FFT (with no additional normalization)
    is :math:`N^3 \times \sigma^2_\mathrm{real}`.

    Furthermore, the power spectrum is defined as V * variance.
    So a normalization factor of 1 / V shows up in step 2,
    cancels this factor such that the power spectrum is P(k).

    The linear displacement field is computed as:

    .. math::

        \psi_i(k) = i \frac{k_i}{k^2} \delta(k)

    .. note::

        To recover the linear velocity in proper units, i.e., km/s,
        from the linear displacement, an additional factor of
        :math:`f \times a \times H(a)` is required

    Parameters
    ----------
    pm : pmesh.pm.ParticleMesh
        the mesh object
    linear_power : callable
        a function taking wavenumber as its only argument, which returns
        the linear power spectrum
    seed : int
        the random seed used to generate the random field
    compute_displacement : bool, optional
        if ``True``, also return the linear Zel'dovich displacement field;
        default is ``False``
    unitary_amplitude : bool, optional
        if ``True``, the seed gaussian has unitary_amplitude.
    inverted_phase: bool, optional
        if ``True``, the phase of the seed gaussian is inverted

    Returns
    -------
    delta_k : ComplexField
        the real-space Gaussian overdensity field
    disp_k : ComplexField or ``None``
        if requested, the Gaussian displacement field
    """
    if not isinstance(seed, numbers.Integral):
        raise ValueError("the seed used to generate the linear field must be an integer")

    # use pmesh to generate random complex white noise field (done in parallel)
    # variance of complex field is unity
    # multiply by P(k)**0.5 to get desired variance
    delta_k = pm.generate_whitenoise(seed, mode='complex', unitary=unitary_amplitude)

    if inverted_phase: delta_k[...] *= -1

    # initialize the displacement fields for (x,y,z)
    if compute_displacement:
        disp_k = [ComplexField(pm) for i in range(delta_k.ndim)]
        for i in range(delta_k.ndim): disp_k[i][:] = 1j
    else:
        disp_k = None

    # volume factor needed for normalization
    norm = 1.0 / pm.BoxSize.prod()

    # iterate in slabs over fields
    slabs = [delta_k.slabs.x, delta_k.slabs]
    if compute_displacement:
        slabs += [d.slabs for d in disp_k]

    # loop over the mesh, slab by slab
    for islabs in zip(*slabs):
        kslab, delta_slab = islabs[:2] # the k arrays and delta slab

        # the square of the norm of k on the mesh
        k2 = sum(kk**2 for kk in kslab)

        # the linear power (function of k)
        power = linear_power((k2**0.5).flatten())

        # multiply complex field by sqrt of power
        delta_slab[...].flat *= (power*norm)**0.5

        # set k == 0 to zero (zero config-space mean)
        zero_idx = k2 == 0.
        delta_slab[zero_idx] = 0.

        # compute the displacement
        if compute_displacement:

            # ignore division where k==0 and set to 0
            with numpy.errstate(invalid='ignore'):
                for i in range(delta_k.ndim):
                    disp_slab = islabs[2+i]
                    disp_slab[...] *= kslab[i] / k2 * delta_slab[...]
                    disp_slab[zero_idx] = 0. # no bulk displacement

    # return Fourier-space density and displacement (which could be None)
    return delta_k, disp_k


def gaussian_real_fields(pm, linear_power, seed,
                unitary_amplitude=False,
                inverted_phase=False, compute_displacement=False):
    r"""
    Make a Gaussian realization of a overdensity field in
    real-space :math:`\delta(x)`.

    If specified, also compute the corresponding linear Zel'dovich
    displacement field :math:`\psi(x)`, which is related to the
    linear velocity field via:

    Notes
    -----
    See the docstring for :func:`gaussian_complex_fields` for the
    steps involved in generating the fields.

    Parameters
    ----------
    pm : pmesh.pm.ParticleMesh
        the mesh object
    linear_power : callable
        a function taking wavenumber as its only argument, which returns
        the linear power spectrum
    seed : int
        the random seed used to generate the random field
    compute_displacement : bool, optional
        if ``True``, also return the linear Zel'dovich displacement field;
        default is ``False``
    unitary_amplitude : bool, optional
        if ``True``, the seed gaussian has unitary_amplitude.
    inverted_phase: bool, optional
        if ``True``, the phase of the seed gaussian is inverted

    Returns
    -------
    delta : RealField
        the real-space Gaussian overdensity field
    disp : RealField or ``None``
        if requested, the Gaussian displacement field
    """
    # make fourier fields
    delta_k, disp_k = gaussian_complex_fields(pm, linear_power, seed,
                            inverted_phase=inverted_phase,
                            unitary_amplitude=unitary_amplitude,
                            compute_displacement=compute_displacement)

    # FFT the density to real-space
    delta = delta_k.c2r()

    # FFT the velocity back to real space
    if compute_displacement:
        disp = [disp_k[i].c2r() for i in range(delta.ndim)]
    else:
        disp = None

    # return density and displacement (which could be None)
    return delta, disp


def lognormal_transform(density, bias=1.):
    r"""
    Apply a (biased) lognormal transformation of the density
    field by computing:

    .. math::

        F(\delta) = \frac{1}{N} e^{b*\delta}

    where :math:`\delta` is the initial overdensity field and the
    normalization :math:`N` is chosen such that
    :math:`\langle F(\delta) \rangle = 1`

    Parameters
    ----------
    density : array_like
        the input density field to apply the transformation to
    bias : float, optional
        optionally apply a linear bias to the density field;
        default is unbiased (1.0)

    Returns
    -------
    toret : RealField
        the real field holding the transformed density field
    """
    toret = density.copy()
    toret[:] = numpy.exp(bias * density.value)
    toret[:] /= numpy.mean(toret)
    return toret


def poisson_sample_to_points(delta, displacement, pm, nbar, bias=1., seed=None, comm=None):
    """
    Poisson sample the linear delta and displacement fields to points.

    The steps in this function:

    #.  Apply a biased, lognormal transformation to the input ``delta`` field
    #.  Poisson sample the overdensity field to discrete points
    #.  Disribute the positions of particles uniformly within the mesh cells,
        and assign the displacement field at each cell to the particles

    Parameters
    ----------
    delta : RealField
        the linear overdensity field to sample
    displacement : list of RealField (3,)
        the linear displacement fields which is used to move the particles
    nbar : float
        the desired number density of the output catalog of objects
    bias : float, optional
        apply a linear bias to the overdensity field (default is 1.)
    seed : int, optional
        the random seed used to Poisson sample the field to points

    Returns
    -------
    pos : array_like, (N, 3)
        the Cartesian positions of each of the generated particles
    displ : array_like, (N, 3)
        the displacement field sampled for each of the generated particles in the
        same units as the ``pos`` array
    """
    if comm is None:
        comm = MPI.COMM_WORLD

    # create a random state with the input seed
    rng = numpy.random.RandomState(seed)

    # apply the lognormal transformation to the initial conditions density
    # this creates a positive-definite delta (necessary for Poisson sampling)
    lagrangian_bias = bias - 1.
    delta = lognormal_transform(delta, bias=lagrangian_bias)

    # mean number of objects per cell
    H = delta.BoxSize / delta.Nmesh
    overallmean = H.prod() * nbar

    # number of objects in each cell (per rank)
    cellmean = delta.value*overallmean
    cellmean = GatherArray(cellmean.flatten(), comm, root=0)

    # rank 0 computes the poisson sampling
    if comm.rank == 0:
        N = rng.poisson(cellmean)
    else:
        N = None

    # scatter N back evenly across the ranks
    counts = comm.allgather(delta.value.size)
    N = ScatterArray(N, comm, root=0, counts=counts).reshape(delta.shape)

    Nlocal = N.sum() # local number of particles
    Ntot = comm.allreduce(Nlocal) # the collective number of particles
    nonzero_cells = N.nonzero() # indices of nonzero cells

    # initialize the mesh of particle positions and displacement
    # this has the shape: (number of dimensions, number of nonzero cells)
    pos_mesh = numpy.empty(numpy.shape(nonzero_cells), dtype=delta.dtype)
    disp_mesh = numpy.empty_like(pos_mesh)

    # generate the coordinates for each nonzero cell
    for i in range(delta.ndim):

        # particle positions initially on the coordinate grid
        pos_mesh[i] = numpy.squeeze(delta.pm.x[i])[nonzero_cells[i]]

        # displacements for each particle
        disp_mesh[i] = displacement[i][nonzero_cells]

    # rank 0 computes the in-cell uniform offsets
    if comm.rank == 0:
        in_cell_shift = numpy.empty((Ntot, delta.ndim), dtype=delta.dtype)
        for i in range(delta.ndim):
            in_cell_shift[:,i] = rng.uniform(0, H[i], size=Ntot)
    else:
        in_cell_shift = None

    # scatter the in-cell uniform offsets back to the ranks
    counts = comm.allgather(Nlocal)
    in_cell_shift = ScatterArray(in_cell_shift, comm, root=0, counts=counts)

    # initialize the output array of particle positions and displacement
    # this has shape: (local number of particles, number of dimensions)
    pos = numpy.zeros((Nlocal, delta.ndim), dtype=delta.dtype)
    disp = numpy.zeros_like(pos)

    # coordinates of each object (placed randomly in each cell)
    for i in range(delta.ndim):
        pos[:,i] = numpy.repeat(pos_mesh[i], N[nonzero_cells]) + in_cell_shift[:,i]
        pos[:,i] %= delta.BoxSize[i]

    # displacements of each object
    for i in range(delta.ndim):
        disp[:,i] = numpy.repeat(disp_mesh[i], N[nonzero_cells])

    return pos, disp



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging, set_options

import os
import pytest
from numpy.testing import assert_allclose, assert_array_equal

setup_logging()

@MPITest([1])
def test_default_columns(comm):
    CurrentMPIComm.set(comm)

    cat = UniformCatalog(nbar=100, BoxSize=1.0)

    # weight column is default
    assert cat['Weight'].is_default

    # override the default value --> no longer default
    cat['Weight'] = 10.
    assert not cat['Weight'].is_default


@MPITest([1, 4])
def test_save(comm):

    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)

    import tempfile
    import shutil

    # initialize an output directory
    if comm.rank == 0:
        tmpfile = tempfile.mkdtemp()
    else:
        tmpfile = None
    tmpfile = comm.bcast(tmpfile)

    # initialize a uniform catalog
    source = UniformCatalog(nbar=0.2e-2, BoxSize=1024., seed=42)

    # add a non-array attrs (saved as JSON)
    source.attrs['empty'] = None

    # save to a BigFile
    source.save(tmpfile, source.columns)

    # assert that no default columns were saved
    datasets = os.listdir(tmpfile)
    assert not any(col in datasets for col in ['Value', 'Selection', 'Weight'])

    # load as a BigFileCatalog
    source2 = BigFileCatalog(tmpfile, attrs={"Nmesh":32})

    # check sources
    for k in source.attrs:
        assert_array_equal(source2.attrs[k], source.attrs[k])

    # check the data
    def allconcat(data):
        return numpy.concatenate(comm.allgather(data), axis=0)
    assert_allclose(allconcat(source['Position']), allconcat(source2['Position']))
    assert_allclose(allconcat(source['Velocity']), allconcat(source2['Velocity']))

    comm.barrier()
    if comm.rank == 0:
        shutil.rmtree(tmpfile)

@MPITest([1, 4])
def test_tomesh(comm):
    CurrentMPIComm.set(comm)

    source = UniformCatalog(nbar=0.2e-2, BoxSize=1024., seed=42)
    source['Weight0'] = source['Velocity'][:, 0]
    source['Weight1'] = source['Velocity'][:, 1]
    source['Weight2'] = source['Velocity'][:, 2]

    mesh = source.to_mesh(Nmesh=128, compensated=True)
    assert_allclose(source['Position'], mesh['Position'])

    mesh = source.to_mesh(Nmesh=128, compensated=True, interlaced=True)
    assert_allclose(source['Position'], mesh['Position'])

    mesh = source.to_mesh(Nmesh=128, weight='Weight0')

    # bad weight name
    with pytest.raises(ValueError):
        mesh = source.to_mesh(Nmesh=128, weight='Weight3')

    # bad window name
    with pytest.raises(ValueError):
        mesh = source.to_mesh(Nmesh=128, weight='Weight0', window='bad_window')

    # missing Nmesh
    with pytest.raises(ValueError):
        mesh = source.to_mesh()

    # missing BoxSize
    BoxSize = source.attrs.pop("BoxSize")
    with pytest.raises(ValueError):
        mesh = source.to_mesh(Nmesh=128)

@MPITest([4])
def test_bad_column(comm):
    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)

    # read a missing column
    with pytest.raises(ValueError):
        data = source.read(['BAD_COLUMN'])

    # read a missing column
    with pytest.raises(ValueError):
        data = source.get_hardcolumn('BAD_COLUMN')

@MPITest([4])
def test_empty_slice(comm):
    CurrentMPIComm.set(comm)

    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)

    # empty slice returns self
    source2 = source[source['Selection']]
    assert source is source2

    # non-empty selection on root only
    if comm.rank == 0:
        sel = source.rng.choice([True, False], size=source.size)
    else:
        sel = numpy.ones(source.size, dtype=bool)

    # this should trigger a full slice
    source2 = source[sel]
    assert source is not source2


@MPITest([4])
def test_slice(comm):
    CurrentMPIComm.set(comm)

    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)

    # slice a subset
    subset = source[:10]
    assert all(col in subset for col in source.columns)
    assert isinstance(subset, source.__class__)
    assert len(subset) == 10
    assert_array_equal(subset['Position'], source['Position'].compute()[:10])

    subset = source[[0,1,2]]
    assert_array_equal(subset['Position'], source['Position'].compute()[[0,1,2]])

    # cannot slice with list of floats
    with pytest.raises(KeyError):
        subset = source[[0.0,1.0,2.0]]

    # missing column
    with pytest.raises(KeyError):
        col = source['BAD_COLUMN']

@MPITest([4])
def test_dask_slice(comm):
    CurrentMPIComm.set(comm)

    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)

    # add a selection column
    index = numpy.random.choice([True, False], size=len(source))
    source['Selection'] = index

    # slice a column with a dask array
    pos = source['Position']
    pos2 = pos[source['Selection']]
    assert_array_equal(pos.compute()[index], pos2.compute())

@MPITest([1, 4])
def test_index(comm):
    CurrentMPIComm.set(comm)

    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)
    r = numpy.concatenate(comm.allgather(source.Index.compute()))
    assert_array_equal(r, range(source.csize))

    source = source.gslice(0, 1000)
    r = numpy.concatenate(comm.allgather(source.Index.compute()))
    assert_array_equal(r, range(source.csize))
    assert source.Index.dtype == numpy.dtype('i8')

@MPITest([1 ,4])
def test_transform(comm):
    cosmo = cosmology.Planck15
    CurrentMPIComm.set(comm)
    data = numpy.ones(100, dtype=[
            ('Position', ('f4', 3)),
            ('Velocity', ('f4', 3))]
            )

    source = ArrayCatalog(data, BoxSize=100, Nmesh=32)

    source['Velocity'] = source['Position'] + source['Velocity']

    source['Position'] = source['Position'] + source['Velocity']

    # Position triggers  Velocity which triggers Position and Velocity
    # which resolves to the true data.
    # so total is 3.
    assert_allclose(source['Position'], 3)

    mesh = source.to_mesh()

@MPITest([1, 4])
def test_getitem_columns(comm):
    CurrentMPIComm.set(comm)

    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)

    # bad column name
    with pytest.raises(KeyError):
        subset = source[['Position', 'BAD_COLUMN']]

    subset = source[['Position']]

    for col in subset:
        assert_array_equal(subset[col].compute(), source[col].compute())

@MPITest([1, 4])
def test_delitem(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)

    # add a test column
    test = numpy.ones(source.size)
    source['test'] = test

    # cannot delete hard coded column
    with pytest.raises(ValueError):
        del source['Position']

    # cannot delete missing column
    with pytest.raises(ValueError):
        del source['BAD_COLUMN']

    assert 'test' in source
    del source['test']
    assert 'test' not in source

def test_columnaccessor():
    from nbodykit.base.catalog import ColumnAccessor
    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)

    c = source['Position']
    truth = c[0].compute()

    assert isinstance(c, ColumnAccessor)
    c *= 10.
    # c is no longer an accessor because it has transformed.
    assert not isinstance(c, ColumnAccessor)

    # thus it is not affecting original.
    assert_array_equal(source['Position'][0].compute(), truth)
    assert_array_equal(c[0].compute(), truth * 10.)

    # inplace modification is still OK
    source['Position'] *= 10
    assert_array_equal(source['Position'][0].compute(), truth * 10)

    # test pretty print.
    assert 'first' in str(source['Position'])
    assert 'last' in str(source['Position'])

    # test circular reference
    new_col = source['Selection']
    assert isinstance(new_col, ColumnAccessor)
    source['Selection2'] = new_col
    assert source['Selection'].catalog is source
    assert source['Selection2'].catalog is source

@MPITest([1, 4])
def test_copy(comm):

    CurrentMPIComm.set(comm)
    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)
    source['TEST'] = 10
    source.attrs['TEST'] = 'TEST'

    # store original data
    data = {}
    for col in source:
        data[col] = source[col].compute()

    # make copy
    copy = source.copy()

    # modify original
    source['Position'] += 100.
    source['Velocity'] *= 10.

    # check data is equal to original
    for col in copy:
        assert_array_equal(copy[col].compute(), data[col])

    # check meta-data
    for k in source.attrs:
        assert k in copy.attrs

    # adding columns to the copy doesn't add to original source
    copy['TEST2'] = 5.0
    assert 'TEST2' not in source

    # make sure attrs are independent.
    source.attrs['foo'] = 123
    assert 'foo' not in copy.attrs

@MPITest([4])
def test_view(comm):
    CurrentMPIComm.set(comm)

    # the CatalogSource
    source = UniformCatalog(nbar=0.2e-3, BoxSize=1024., seed=42)
    source['TEST'] = 10.
    source.attrs['TEST'] = 10.0

    # view
    view = source.view()
    assert view.base is source
    assert isinstance(view, source.__class__)

    # check meta-data
    for k in source.attrs:
        assert k in view.attrs

    # adding columns to the view changes original source
    view['TEST2'] = 5.0
    assert 'TEST2' in source

    # make sure attrs are dependent.
    source.attrs['foo'] = 123
    assert 'foo' in view.attrs
